/***************************************************************************
 *                                                                         *     
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/    

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>

#include "Constants.h"
#include "ControlCentre.h"
#include "DBElement.h"
#include "DB.h"
#include "Socket.h"
#include "Inet.h"
#include "GnutellaPacket.h"

#ifndef ROUTER
#include "SearchResult.h"
#endif

#include "Packet.h"
#include "Mesg.h"
#include "Buffer.h"
#include "Connection.h"
#include "GnutellaConnection.h"
#include "Host.h"

#include "FSEntry.h"
#include "Path.h"
#include "File.h"
#include "Transfer.h"
#include "TransferStatus.h"
#include "Upload.h"
#include "LocalResult.h"

#ifndef ROUTER
#include "Download.h"
#include "HTTPClient.h"
#include "SearchManager.h"
#include "Downloader.h"
#endif

#include "HTTPServer.h"
#include "FSScanner.h"  
#include "FServer.h"
#include "Networker.h"
#include "GnutellaRouter.h"
#include "HostCatcher.h"
#ifndef ROUTER
#include "Query.h"
#include "Monitor.h"
#endif
#include "Backend.h"


/*
 * General stuff
 */

Backend::Backend(ControlCentre *newMCC)
{
	MCC		= newMCC;
	connections	= new DB(MCC->get_MAX_CONNECTIONS(), HASH_TINY);
	hosts		= new DB(MCC->get_MAX_HOSTS(), HASH_SMALL);
	fileServer	= new FServer(MCC);
#ifndef ROUTER
	searchMan	= new SearchManager(MCC);
	leechMan	= new Downloader(MCC);
	bigBrother	= new Monitor(MCC);
#endif
	myHostCatcher	= new HostCatcher(MCC);
	GRouter		= new GnutellaRouter(MCC);
	netMan		= new Networker(MCC);
	hostIndex	= 0;
	ourInit		= new Mesg();
#ifndef ROUTER
	sign		= 0;
	hostsOnNet	= 0;
	filesOnNet	= 0;
	kbsOnNet	= 0;
#endif
}

int Backend::reload()
{
	return 1;
}

Backend::~Backend()
{
	Connection	*c;
	int		i;

#ifndef ROUTER
	delete(searchMan);
	if (sign)
		delete(sign);
	delete(ourInit);
#endif
	hosts->clearDB();
	delete(hosts);
	delete(GRouter);
	delete(netMan);

	for (i = 0; i < connections->getSize(); i++)
		if ((c = (Connection *)connections->findElement(i)))
			c->close();

	connections->clearDB();
	delete(connections);
}


/*
 * GnutellaNet interface
 */

int Backend::pollGnutellaNet()
{
	Connection	*c;
	Connection	*cnext;
	GnutellaPacket	*p;
	GnutellaPacket	*prv;
	int		res;
	int		i;

	res	= autoOpenConnections();

	/*
	 * Compare Packet exchange ratio to kill off too slow and too fast people.
	 * Add 150 packets to reduce effects of small numbers
	 */
	if (MCC->get_PACKET_RATIO())
		for (i = 0; i < connections->getSize(); i++)
			if (	((c = (Connection *)connections->findElement(i)))	&&
				(c->getType() == CON_TYPE_GNUTELLA)			)
			{
				float	ratio;

				ratio	=	((float)((GnutellaConnection *)c)->getRecvd() + 150)	/
					//-------------------------------------------------------------
						((float)((GnutellaConnection *)c)->getSent() + 150 );

				if ( (100 * ratio < MCC->get_PACKET_RATIO())	||
							(500 / ratio < MCC->get_PACKET_RATIO())	)
 					closeConnection(i);
			}

#ifndef ROUTER
	if (autoPush())
		res	= 1;

	res	= 	( leechMan->pollDownloader() || res );
#endif
	res	=	( fileServer->pollFServer() || res );

	c	= netMan->waitConnection(hosts);

	res	= (c || res);

	while (c)
	{
		cnext	= (Connection *)c->getLeft();
		handleNewConnection(c);
		c	= cnext;
	}

#ifndef ROUTER
	if (handlePendingOutgoingConnections())
		res	= 1;
#endif

	p	= netMan->waitTraffic(connections);
	if (p)
	{
		while (p)
		{
			if (handleNewPacket(p))
				GRouter->handle(p, connections);
			prv	= p;
			p	= (GnutellaPacket *)p->getLeft();
			delete(prv);
		}
		res	= 1;
	}

	res	= (netMan->doSend(connections) || res);

	return res;
}

int Backend::startListening()
{
	return netMan->startListening();
}

int Backend::sendInit()
{
	GnutellaPacket	*p;
	int		res;

	p	= new GnutellaPacket(FUNC_INIT, 0, 0, MCC->get_DEFAULT_TTL());
	p->setMesgID((unsigned char *)ourInit->getMesgIDp());
	res	= GRouter->broadcast(p, connections);
	delete p;
	return res;
}

#ifndef ROUTER

unsigned int Backend::getHostsOnNet()
{
	return hostsOnNet;
}

unsigned long int Backend::getFilesOnNet()
{
	return filesOnNet;
}

unsigned long int Backend::getSizeOnNet()
{
	return kbsOnNet / 1024;
}


/*
 * Monitor interface
 */

Query *Backend::getQueries()
{
	return bigBrother->getQueries();
}

unsigned int Backend::getQueryCount()
{
	return bigBrother->getQueryCount();
}
#endif

/*
 * FileServer interface
 */

int Backend::addPath(char *newPath)
{
	return fileServer->addPath(newPath);
}

int Backend::delPath(char *oldPath)
{
	return fileServer->deletePath(oldPath);
}

int Backend::delAllPaths()
{
	return fileServer->delAllPath();
}

int Backend::scanDirs()
{
	return fileServer->doScan();
}

int Backend::killUpload(int idx)
{
	return fileServer->killUpload(idx);
}

TransferStatus *Backend::getUploads()
{
	return fileServer->getStatusList();
}

Path *Backend::getPathList()
{
	return fileServer->getPaths();
}

int Backend::getFilesShared()
{
	return fileServer->getFilesCount();
}

int Backend::getSizeShared()
{
	return fileServer->getSizeShared();
}

unsigned int Backend::getLocalSearches()
{
	return fileServer->getLocalSearches();
}

unsigned int Backend::getCacheHits()
{
	return fileServer->getCacheHits();
}

unsigned int Backend::getCacheEntries()
{
	return fileServer->getCacheEntries();
}

unsigned int Backend::getCacheSize()
{
	return fileServer->getCacheSize();
}

unsigned int Backend::getSearchArraySize()
{
	return fileServer->getSearchArraySize();
}

unsigned int Backend::getHashArraySize()
{
	return fileServer->getHashArraySize();
}


#ifndef ROUTER

/*
 * SearchManager interface
 */

int Backend::launchQuery(int speed, char *findMe)
{
	Packet	*p;
	int	res;

	if (sign)
		delete(sign);
	p	= new Packet(MCC->get_DEFAULT_TTL());
	p->mkSearchPacket(speed, findMe, MCC->get_DEFAULT_TTL());
	sign	= new Mesg(p);
	searchMan->newQuery(speed, sign);
	res	= GRouter->broadcast(p, connections);
	delete p;

	return res;
}

SearchResult *Backend::getOneResult(int i)
{
	return searchMan->getSearchResult(i);
}

void Backend::clearResults()
{
	searchMan->clearResults();
}

unsigned int Backend::getSMQueries()
{
	return searchMan->getQueries();
}

unsigned int Backend::getSMResultCount()
{
	return searchMan->getResultCount();
}

/*
 * Download interface
 */

int Backend::startDownload(int idx)
{
	SearchResult	*s;

	if (!(s = searchMan->getSearchResult(idx)))
	{
		MCC->addError("result not found!");
		return 0;
	}

	return leechMan->startDownload(s->getHostIP(), s->getHostPort(),
			s->getFileIndex(), s->getNamep(), s->getGClientIDp(),
			s->getConnectionIndex());
}

int Backend::continueDownload(int idx)
{
	SearchResult	*s;

	if (!(s = searchMan->getSearchResult(idx)))
	{
		MCC->addError("result not found!");
		return 0;
	}

	return leechMan->continueDownload(s->getHostIP(), s->getHostPort(),
			s->getFileIndex(), s->getNamep(), s->getGClientIDp(),
			s->getConnectionIndex());
}

int Backend::killDownload(int idx)
{
	return leechMan->killDownload(idx);
}

int Backend::pushDownload(int idx)
{
	SearchResult		*s;
	Packet			*p;
	int			i;
	char			host[257];

	if (!(s = searchMan->getSearchResult(idx)))
	{
		MCC->addError("No such result");
		return 0;
	}

	if (!Inet::getHostName(host, 256))
	{
		MCC->addError("gethostname failed");
		return 0;
	}

	p	= new Packet(MCC->get_DEFAULT_TTL());
	p->mkPushPacket(s->getGClientIDp(), s->getFileIndex(), Inet::NSLookup(host), MCC->get_PORT(), MCC->get_DEFAULT_TTL());
	p->setTargetConnection(s->getConnectionIndex());
	if (!leechMan->newPushRequest(s->getFileIndex(), s->getNamep(),
						p->getMesgIDp(), s->getGClientIDp(), 0))
 	{
		delete p;
		return 0;
	}

	i	= GRouter->sendPacket(p, connections);
	delete p;

	return i;
}

TransferStatus *Backend::getDownloads()
{
	return leechMan->getStatusList();
}

unsigned int Backend::getDLBytesDownloaded()
{
	return leechMan->getBytesDownloaded();
}
#endif

/*
 * Auto things
 */

void Backend::set_AUTO_CONNECTIONS(int nc)
{
	MCC->set_AUTO_CONNECTIONS(nc);
}

void Backend::set_AUTO_PUSH(int yn)
{
	MCC->set_AUTO_PUSH(yn);
}

void Backend::set_AUTO_CLEAR_COMPLETE(int yn)
{
	MCC->set_AUTO_CLEAR_COMPLETE(yn);
}

/*
 * Controls
 */

int Backend::get_INT_PARAM(int which)
{
	switch(which)
	{
		case PARAM_MAX_TTL	:
				return MCC->get_MAX_TTL();
		case PARAM_DEFAULT_TTL	:
				return MCC->get_DEFAULT_TTL();
		case PARAM_ROUTING_TABLE_SIZE	:
				return MCC->get_ROUTING_TABLE_SIZE();
		case PARAM_MAX_CONNECTIONS	:
				return MCC->get_MAX_CONNECTIONS();
		case PARAM_MAX_RESULTS	:
				return MCC->get_MAX_RESULTS();
		case PARAM_MAX_HOPS	:
				return MCC->get_MAX_HOPS();
		case PARAM_MAX_UPLOADS	:
				return MCC->get_MAX_UPLOADS();
		case PARAM_MAX_DOWNLOADS	:
				return MCC->get_MAX_DOWNLOADS();
		case PARAM_PORT	:
				return MCC->get_PORT();
		case PARAM_SPEED	:
				return MCC->get_SPEED();
		case PARAM_AUTO_CLEAR_COMPLETE	:
				return MCC->get_AUTO_CLEAR_COMPLETE();
		case PARAM_AUTO_CONNECTIONS	:
				return MCC->get_AUTO_CONNECTIONS();
		case PARAM_AUTO_PUSH	:
				return MCC->get_AUTO_PUSH();
		case PARAM_AUTO_CLEAR_RESULTS	:
				return MCC->get_AUTO_CLEAR_RESULTS();
		case PARAM_MAX_HOST_CON_FAIL	:
				return MCC->get_MAX_HOST_CON_FAIL();
		case PARAM_MAX_MONITOR_LENGTH	:
				return MCC->get_MAX_MONITOR_LENGTH();
		case PARAM_MAX_FILES_SHARED	:
				return MCC->get_MAX_FILES_SHARED();
		case PARAM_MAX_CON_ERRORS	:
				return MCC->get_MAX_CON_ERRORS();
		case PARAM_IGNORE_LOCAL_IP	:
				return MCC->get_IGNORE_LOCAL_IP();
		case PARAM_NOTIFY_RESULTS	:
				return MCC->get_NOTIFY_RESULTS();
		case PARAM_CACHE_ENTRIES	:
				return MCC->get_CACHE_ENTRIES();
		case PARAM_PACKET_RATIO		:
				return MCC->get_PACKET_RATIO();
		default	:	MCC->addError("Backend::get_INT_PARAM : "
						"unknown parameter!!!");
	}
	return -1;
}

void Backend::get_STR_PARAM(int which, char *buf)
{
	switch(which)
	{
		case PARAM_CLIENTID	:
				memcpy(buf, ourClientID, CLIENTID_LENGTH);
				break;
		case PARAM_DOWNLOAD_DIR	:
				MCC->get_DOWNLOAD_DIR(buf);
				break;
		default	:	MCC->addError("Backend::get_STR_PARAM : "
						"Unknwon parameter!!!");
	}
}

void Backend::set_INT_PARAM(int which, int val)
{
	DB		*newConnections;
	int		i;

	switch (which)
	{
		case PARAM_MAX_TTL	:
				MCC->set_MAX_TTL(val);
				break;
		case PARAM_DEFAULT_TTL	:
				MCC->set_DEFAULT_TTL(val);
				break;
		case PARAM_MAX_CONNECTIONS	:
				if (val == MCC->get_MAX_CONNECTIONS())
					return;
				MCC->set_MAX_CONNECTIONS(val);
				newConnections	= new DB(val, HASH_TINY);
				for (i = 0; i < val; i++)
				{
					Connection	*c;

					if ((c = (Connection *)connections->findElement(i)))
						newConnections->storeElement(c);
				}
				newConnections->balance();
				delete(connections);
				connections	= newConnections;
				break;
		case PARAM_ROUTING_TABLE_SIZE	:
				if (val == MCC->get_ROUTING_TABLE_SIZE())
					break;
				MCC->set_ROUTING_TABLE_SIZE(val);
				GRouter->resize();
				break;
		case PARAM_MAX_HOPS	:
				MCC->set_MAX_HOPS(val);
				break;
		case PARAM_MAX_UPLOADS	:
				MCC->set_MAX_UPLOADS(val);
				fileServer->resize();
				break;
		case PARAM_PORT		:
				netMan->stopListening();
				MCC->set_PORT(val);
				netMan->startListening();
				break;
		case PARAM_SPEED	:
				MCC->set_SPEED(val);
				break;
		case PARAM_AUTO_PUSH	:
				MCC->set_AUTO_PUSH(val);
				break;
		case PARAM_AUTO_SAVE_HOSTS	:
				MCC->set_AUTO_SAVE_HOSTS(val);
				break;
		case PARAM_AUTO_CONNECTIONS	:
				MCC->set_AUTO_CONNECTIONS(val);
				break;
		case PARAM_AUTO_CLEAR_COMPLETE	:
				MCC->set_AUTO_CLEAR_COMPLETE(val);
				break;
		case PARAM_AUTO_CLEAR_RESULTS	:
				MCC->set_AUTO_CLEAR_RESULTS(val);
				break;
		case PARAM_MAX_HOST_CON_FAIL	:
				MCC->set_MAX_HOST_CON_FAIL(val);
				break;
		case PARAM_MAX_FILES_SHARED	:
				if (val == MCC->get_MAX_FILES_SHARED())
					break;
				MCC->set_MAX_FILES_SHARED(val);
				fileServer->doScan();
				break;
		case PARAM_MAX_CON_ERRORS	:
				MCC->set_MAX_CON_ERRORS(val);
				break;
#ifndef ROUTER
		case PARAM_MAX_MONITOR_LENGTH	:
				if (val == MCC->get_MAX_MONITOR_LENGTH())
					break;
				MCC->set_MAX_MONITOR_LENGTH(val);
				bigBrother->resize();
				break;
		case PARAM_MAX_RESULTS	:
				MCC->set_MAX_RESULTS(val);
				leechMan->resize();
				break;
		case PARAM_MAX_DOWNLOADS	:
				if (val == MCC->get_MAX_DOWNLOADS())
					break;
				MCC->set_MAX_DOWNLOADS(val);
				MCC->set_MAX_PUSH_REQUESTS(val);
				leechMan->resize();
				break;
#endif
		case PARAM_IGNORE_LOCAL_IP	:
				MCC->set_IGNORE_LOCAL_IP(val);
				break;
		case PARAM_NOTIFY_RESULTS	:
				MCC->set_NOTIFY_RESULTS(val);
				break;
		case PARAM_CACHE_ENTRIES	:
				if (val == MCC->get_CACHE_ENTRIES())
					break;
				MCC->set_CACHE_ENTRIES(val);
				fileServer->resizeCache();
				break;
		case PARAM_PACKET_RATIO	:
				MCC->set_PACKET_RATIO(val);
				break;
		default	: MCC->addError("Backend::set_INT_PARAM : Unknown "
					"parameter!!!");
	}
}

void Backend::set_STR_PARAM(int which, char *buf)
{
	switch (which)
	{
		case PARAM_CLIENTID	:
				memcpy(ourClientID, buf, CLIENTID_LENGTH);
				MCC->set_CLIENTID(buf);
				break;
		case PARAM_DOWNLOAD_DIR	:
				MCC->set_DOWNLOAD_DIR(buf);
				break;
		default	:	MCC->addError("Backend::set_STR_PARAM : "
						"unknown parameter!!!");
	}
}

int Backend::deleteHost(int i)
{
	if (
#ifndef ROUTER
	 	leechMan->isUsingHost(i)	||
#endif
		fileServer->isUsingHost(i)	)
	{
		MCC->addError("Host is in use!");
		return 0;
	}

        if (i >= 0)
	{
		closeHost(i);
		return hosts->deleteElement(i);
	}

        return 0;
}

int Backend::storeHost(char *dotted, int tries, char *cid)
{
	Host	*h;
	int	res;

	h	= new Host(dotted, cid);
	h->setTries(tries);

	if ((res = hosts->storeElement(h)) == ENTRY_INVALID)
		delete h;

	return res;
}

int Backend::closeConnection(int i)
{
	Connection	*c;

	c	= (Connection *)connections->findElement(i);
	if (c)
	{
		Host	*h;
		
		c->close();
		if (!((h = (Host *)hosts->findElement(c->getHostIndex()))))
			MCC->addError("BUG : Backend::closeConnection() : can't find host");
		else
			if (!h->isReal())
				hosts->deleteElement(c->getHostIndex());
		
		if (!connections->deleteElement(i))
		{
			MCC->addError("BUG : Backend : couldn't delete connection");
			return 0;
		}
		return 1;
	}
	return 0;
}

int Backend::closeHost(int i)
{
	Connection	*c;
	int		j;
	int		cnt;

	cnt	= 0;
	for (j = 0; j < connections->getSize(); j++)
	{
		c	= (Connection *)connections->findElement(j);
		if (c)
			if (c->getHostIndex() == i)
			{
				c->close();
				connections->deleteElement(j);
				cnt++;
			}
	}
	return cnt;
}

#ifndef ROUTER

int Backend::handlePendingOutgoingConnections()
{
	Packet  *p;
	Connection	*c;
        char            feedback[30];
        int             res;
	int		i;

	res	= 0;

	for(i = 0; i < connections->getSize(); i++)
		if ((c = (Connection *)connections->findElement(i)))
		{
			Host	*h;

			if (!c->isOutgoing())
				continue;
			res		= 1;
			switch (c->getState())
			{
				case	CON_STATE_INACTIVE	:
				case	CON_STATE_INIT		:
					if (c->tryToConnect() == -1)
					{
						if ((h = (Host *)hosts->findElement(c->getHostIndex())))
						{
							h->incTries();
							if (h->getTries() >= MCC->get_MAX_HOST_CON_FAIL())
								deleteHost(h->getEntry());
						}
						else
							MCC->addError("Backend : Host not "
								"found in handlePendingOutgo"
								"ingConnections()!!!");
					}
					break;
				case	CON_STATE_CONNECTED	:
					if (c->sendStr("GNUTELLA CONNECT/0.4\n\n",
						strlen("GNUTELLA CONNECT/0.4\n\n"))
							< (int)strlen("GNUTELLA CONNECT/0.4\n\n"))
					{
						c->setState(CON_STATE_LOGIN_FAILED);
						if ((h = (Host *)hosts->findElement(c->getHostIndex())))
						{
							h->incTries();
							if (h->getTries() >= MCC->get_MAX_HOST_CON_FAIL())
								deleteHost(h->getEntry());
 						}
						else
							MCC->addError("Backend : Host not "
								"found in handlePendingOutgo"
								"ingConnections()!!!");
					}
					else
					{
						c->setState(CON_STATE_LOGIN_SENT);
						c->bufferSend();
					}
					break;
				case	CON_STATE_LOGIN_SENT	:
					if (c->getOutBufferBytes() > 0)
					{
						if (c->getAge()	> 30)
							c->setState(CON_STATE_LOGIN_FAILED);
						else
							c->bufferSend();
						break;
					}
					if (c->read() == -1)
					{
						c->setState(CON_STATE_CON_ERROR);
						break;
					}
					if (c->getInBufferBytes() < strlen("GNUTELLA OK\n\n"))
					{
						if (c->getAge() > CONNECTION_TIMEOUT)
						{
							c->setState(CON_STATE_LOGIN_FAILED);
							if ((h = (Host *)hosts->findElement(c->getHostIndex())))
							{
								h->incTries();
								if (h->getTries() >= MCC->get_MAX_HOST_CON_FAIL())
									deleteHost(h->getEntry());
							}
							else
								MCC->addError("Backend : Host not "
									"found in handlePendingOutgo"
									"ingConnections()!!!");
						}
						break;
					}
					c->getInBuffer(feedback, strlen("GNUTELLA OK\n\n"));
					feedback[ strlen("GNUTELLA OK\n\n") ]	= 0;
					if (strcmp(feedback, "GNUTELLA OK\n\n"))
					{
						c->setState(CON_STATE_LOGIN_FAILED);
						if ((h = (Host *)hosts->findElement(c->getHostIndex())))
						{
							h->incTries();
							if (h->getTries() >= MCC->get_MAX_HOST_CON_FAIL())
								deleteHost(h->getEntry());
						}
						else
							MCC->addError("Backend : Host not "
								"found in handlePendingOutgo"
								"ingConnections()!!!");
						break;
					}
					c->skipBytes(strlen("GNUTELLA OK\n\n"));
					c->setState(CON_STATE_ACTIVE);
					if ((h = (Host *)hosts->findElement(c->getHostIndex())))
						h->setTries(0);
					else
						MCC->addError("Backend : Host not "
							"found in handlePendingOutgo"
							"ingConnections()!!!");
	        			p       = new Packet(MCC->get_DEFAULT_TTL());
					p->mkInitPacket(MCC->get_DEFAULT_TTL());
					p->setMesgID((unsigned char *)ourInit->getMesgIDp());
					p->setTargetConnection(i);
					GRouter->sendPacket(p, connections);
	       	 			delete p;
					break;
				case CON_STATE_ACTIVE	: continue;
				default	:
					if ((h = (Host *)hosts->findElement(c->getHostIndex())))
					{
						h->incTries();
						if (h->getTries() >= MCC->get_MAX_HOST_CON_FAIL())
							deleteHost(h->getEntry());
					}
					else
						MCC->addError("Backend : Host not "
							"found in handlePendingOutgo"
							"ingConnections()!!!");
					closeConnection(i);
					break;
			}
		}

        return res;
}

#endif

int Backend::connectToHost(int i)
{
        Host    	        *h;
        GnutellaConnection      *gc;

	if (connections->isFull())
		return 0;

        if (!(h = (Host *)hosts->findElement(i)))
        {
                MCC->addError("HOST NOT FOUND!!!");
                return 0;
        }

	gc	= new GnutellaConnection();
	if (!((gc->Connect(h->getNetIP(), h->getPort()))))
	{
		h->incTries();
		if (h->getTries() >= MCC->get_MAX_HOST_CON_FAIL())
			deleteHost(h->getEntry());
		delete gc;
		return 0;
	}

	gc->setHostIndex(h->getEntry());
	gc->setType(CON_TYPE_GNUTELLA);
	if (connections->storeElement(gc) == ENTRY_INVALID)
	{
		MCC->addError("Couldn't store Connection!");
		gc->close();
		delete gc;
		return 0;
	}
	
	return 1;
}

GnutellaConnection *Backend::getConnection(int i)
{
	return (GnutellaConnection *)connections->findElement(i);
}

Host *Backend::getHost(int i)
{
	return (Host *)hosts->findElement(i);
}

#ifndef ROUTER

int Backend::writeOptions(char *fileName)
{
	FILE	*file;
	int	i;
	char	sparameter[FILE_NAME_LENGTH];

	if (!(file = fopen(fileName, "w")))
		return 0;

	for (i = 0; i < PARAM_INVALID; ++i)
	{
		switch(i)
		{
			case PARAM_MAX_TTL :
				if (MCC->get_MAX_TTL() != DEFAULT_MAX_TTL)
					fprintf(file, "MAX_TTL %d\n", MCC->get_MAX_TTL());
				break;
			case PARAM_DEFAULT_TTL :
				if (MCC->get_DEFAULT_TTL() != DEFAULT_DEFAULT_TTL)
					fprintf(file, "DEFAULT_TTL %d\n", MCC->get_DEFAULT_TTL());
				break;
			case PARAM_ROUTING_TABLE_SIZE :
				if (MCC->get_ROUTING_TABLE_SIZE() != DEFAULT_ROUTING_TABLE_SIZE)
					fprintf(file, "ROUTING_TABLE_SIZE %d\n",
							MCC->get_ROUTING_TABLE_SIZE());
				break;
			case PARAM_MAX_CONNECTIONS :
				if (MCC->get_MAX_CONNECTIONS() != DEFAULT_MAX_CONNECTIONS)
					fprintf(file, "MAX_CONNECTIONS %d\n", MCC->get_MAX_CONNECTIONS());
				break;
			case PARAM_MAX_RESULTS :
				if (MCC->get_MAX_RESULTS() != DEFAULT_MAX_RESULTS)
					fprintf(file, "MAX_RESULTS %d\n", MCC->get_MAX_RESULTS());
				break;
			case PARAM_MAX_HOPS :
				if (MCC->get_MAX_HOPS() != DEFAULT_MAX_HOPS)
					fprintf(file, "MAX_HOPS %d\n", MCC->get_MAX_HOPS());
				break;
			case PARAM_MAX_PUSH_REQUESTS :
				if (MCC->get_MAX_PUSH_REQUESTS() != DEFAULT_MAX_PUSH_REQUESTS)
					fprintf(file, "MAX_PUSH_REQUESTS %d\n",
						MCC->get_MAX_PUSH_REQUESTS());
				break;
			case PARAM_MAX_UPLOADS :
				if (MCC->get_MAX_UPLOADS() != DEFAULT_MAX_UPLOADS)
					fprintf(file, "MAX_UPLOADS %d\n", MCC->get_MAX_UPLOADS());
				break;
			case PARAM_MAX_DOWNLOADS :
				if (MCC->get_MAX_DOWNLOADS() != DEFAULT_MAX_DOWNLOADS)
					fprintf(file, "MAX_DOWNLOADS %d\n", MCC->get_MAX_DOWNLOADS());
				break;
			case PARAM_MAX_HOSTS :
				if (MCC->get_MAX_HOSTS() != DEFAULT_MAX_HOSTS)
					fprintf(file, "MAX_HOSTS %d\n", MCC->get_MAX_HOSTS());
				break;
			case PARAM_PORT :
				if (MCC->get_PORT() != DEFAULT_PORT)
					fprintf(file, "PORT %d\n", MCC->get_PORT());
				break;
			case PARAM_SPEED :
				if (MCC->get_SPEED() != DEFAULT_SPEED)
					fprintf(file, "SPEED %d\n", MCC->get_SPEED());
				break;
			case PARAM_MAX_MONITOR_LENGTH :
				if (MCC->get_MAX_MONITOR_LENGTH() != DEFAULT_MAX_MONITOR_LENGTH)
					fprintf(file, "MAX_MONITOR_LENGTH %d\n",
				MCC->get_MAX_MONITOR_LENGTH());
				break;
			case PARAM_AUTO_CONNECTIONS :
				if (MCC->get_AUTO_CONNECTIONS() != DEFAULT_AUTO_CONNECTIONS)
					fprintf(file, "AUTO_CONNECTIONS %d\n", MCC->get_AUTO_CONNECTIONS());
				break;
			case PARAM_AUTO_PUSH :
				if (MCC->get_AUTO_PUSH() != DEFAULT_AUTO_PUSH)
					fprintf(file, "AUTO_PUSH %d\n", MCC->get_AUTO_PUSH());
				break;
			case PARAM_AUTO_CLEAR_COMPLETE :
				if (MCC->get_AUTO_CLEAR_COMPLETE() != DEFAULT_AUTO_CLEAR_COMPLETE)
					fprintf(file, "AUTO_CLEAR_COMPLETE %d\n",
				MCC->get_AUTO_CLEAR_COMPLETE());
				break;
			case PARAM_AUTO_CLEAR_RESULTS :
				if (MCC->get_AUTO_CLEAR_RESULTS() != DEFAULT_AUTO_CLEAR_RESULTS)
					fprintf(file, "AUTO_CLEAR_RESULTS %d\n",
				MCC->get_AUTO_CLEAR_RESULTS());
				break;
			case PARAM_MAX_HOST_CON_FAIL :
				if (MCC->get_MAX_HOST_CON_FAIL() != DEFAULT_MAX_HOST_CON_FAIL)
					fprintf(file, "MAX_HOST_CON_FAIL %d\n",
						MCC->get_MAX_HOST_CON_FAIL());
				break;
			case PARAM_AUTO_SAVE_HOSTS	:
				if (MCC->get_AUTO_SAVE_HOSTS() != DEFAULT_AUTO_SAVE_HOSTS)
					fprintf(file, "AUTO_SAVE_HOSTS %d\n",
						MCC->get_AUTO_SAVE_HOSTS());
				break;
			case PARAM_MAX_FILES_SHARED :
				if (MCC->get_MAX_FILES_SHARED() != DEFAULT_MAX_FILES_SHARED)
					fprintf(file, "MAX_FILES_SHARED %d\n",
						MCC->get_MAX_FILES_SHARED());
				break;
			case PARAM_MAX_CON_ERRORS :
				if (MCC->get_MAX_CON_ERRORS() != DEFAULT_MAX_CON_ERRORS)
					fprintf(file, "MAX_CON_ERRORS %d\n",
						MCC->get_MAX_CON_ERRORS());
				break;
			case PARAM_IGNORE_LOCAL_IP :
				if (MCC->get_IGNORE_LOCAL_IP() != DEFAULT_IGNORE_LOCAL_IP)
					fprintf(file, "IGNORE_LOCAL_IP %d\n",
						MCC->get_IGNORE_LOCAL_IP());
				break;
			case PARAM_NOTIFY_RESULTS :
				if (MCC->get_NOTIFY_RESULTS() != DEFAULT_NOTIFY_RESULTS)
					fprintf(file, "NOTIFY_RESULTS %d\n",
						MCC->get_NOTIFY_RESULTS());
				break;
			case PARAM_CACHE_ENTRIES :
				if (MCC->get_CACHE_ENTRIES() != DEFAULT_CACHE_ENTRIES)
					fprintf(file, "CACHE_ENTRIES %d\n",
						MCC->get_CACHE_ENTRIES());
				break;
			case PARAM_PACKET_RATIO	:
				if (MCC->get_PACKET_RATIO() != DEFAULT_PACKET_RATIO)
					fprintf(file, "PACKET_RATIO %d\n",
						MCC->get_PACKET_RATIO());
				break;
			case PARAM_DOWNLOAD_DIR :
				MCC->get_DOWNLOAD_DIR(sparameter);
				fprintf(file, "DOWNLOAD_DIR %s\n", sparameter);
				break;
			case PARAM_DAEMON_PORT :
#ifdef DAEMON
				if (MCC->get_DAEMON_PORT() != DEFAULT_DAEMON_PORT)
					fprintf(file, "DAEMON_PORT %d\n", MCC->get_DAEMON_PORT());
#endif
				break;
			case PARAM_CLIENTID	:
				break;
			default : MCC->addError("Backend::writeOptions : "
						"Unknown option!!");
				break;
		}
	}

	fclose(file);
	return 1;
}

#endif

int Backend::writeShares(char *fileName)
{
        int     i;
        Path	*p;
	FILE	*f;

        i    = 0;
        if ((f = fopen(fileName, "w")))
	{
		for (p = fileServer->getPaths(); p; p = (Path *)p->getNext())
			if (!p->isSubdir())
			{
				fprintf(f, "%s\n", p->getPathp());
				i++;
			}
		fclose(f);
	}

	return i;
}

int Backend::writeHosts(char *fileName)
{
        int     i;
        int     fcnt;
        char    host[DOTTED_LENGTH];
	char	cid[CLIENTID_LENGTH];
        Host    *h;
        FILE    *f;

        if ((f = fopen(fileName, "w")))
        {
	        fcnt    = 0;
                for (i = 0; i < hosts->getSize(); i++)
                        if ((h  = (Host *)hosts->findElement(i)))
                        {
				if (!h->isReal())
					continue;

				if (h->getTries() > MCC->get_MAX_HOST_CON_FAIL())
					continue;

                                h->getDotted(host);
				if (h->hasClientID())
				{
					int	i;

					h->getClientID(cid);
	                                fprintf(f, "%s %d CID: ", host,
							h->getTries());
					for(i = 0; i < CLIENTID_LENGTH; i++)
						fprintf(f, "%c", cid[i]);
					fprintf(f, "\n");
				}
				else
	                                fprintf(f, "%s %d NOCID\n", host,
							h->getTries());
                                fcnt++;
                        }
                fclose(f);
	        return fcnt;
        }
	else
	{
		MCC->addError("Backend::writeHosts : Can't open file");
		return -1;
	}
}

int Backend::handleNewPacket(GnutellaPacket *p)
{
	Mesg	*m;
	
	if (!p->isSane())
	{
		GnutellaConnection	*c;
		
		if(!((c = (GnutellaConnection *)connections->findElement(p->getSourceConnection()))))
			return 0;

		c->punchErrors();
		c->punchDropped();
		if (c->getErrors() > MCC->get_MAX_CON_ERRORS())
			closeConnection(c->getEntry());
		return 0;
	}

	m	= new Mesg(p);
	if (GRouter->alreadySeen(m) && (!(p->getType() & 0x1)))
	{
		GnutellaConnection	*c;
		
		if ((c = (GnutellaConnection *)connections->findElement(p->getSourceConnection())))
			c->punchDropped();

		delete m;
		return 0;
	}
	delete m;
	switch (p->getType())
	{
		case FUNC_INIT		: return handleInit(p);
		case FUNC_INIT_RESP	: myHostCatcher->catchHost(p, hosts);
#ifndef ROUTER
					m	= new Mesg(p);
					if (m->equals(ourInit))
					{
						delete m;
						handleInitResp(p);
						return 0;
					}
					else
					{
						delete m;
						return 1;
					}
#else
					return 1;
#endif
		case FUNC_PUSH	: return handlePush(p);
		case FUNC_SEARCH	: handleSearch(p);
					return 1;
		case FUNC_SEARCH_RESP	: myHostCatcher->catchHost(p, hosts);
#ifndef ROUTER	
					m	= new Mesg(p);
					if (searchMan->getSignature())
					{
						if (m->equals(searchMan->getSignature()))
						{
	                                     		searchMan->addResults(p);
							delete m;
							return 0;
						}
						else
						{
							delete m;
							return 1;
						}
					}
					delete m;
					return 1;
#else
					return 1;
#endif
					break;
//		case FUNC_	:
		default		: break;
	}

	// Unknown Packet ? lets drop it.
	return 0;
}

int Backend::handleInit(GnutellaPacket *p)
{
	Packet			*resp;
	GnutellaConnection	*c;
	char			host[256];

	if (!(c = (GnutellaConnection *)connections->findElement(p->getSourceConnection())))
		return 0;

	if (!Inet::getHostName(host, 256))
		return 0;

	resp	= new Packet(p);
	if (resp->mkInitRespPacket(MCC->get_PORT(), Inet::NSLookup(host),
		fileServer->getFilesCount(), (fileServer->getSizeShared() / 1024), MCC->get_DEFAULT_TTL()))
	{
		resp->setMesgID((unsigned char *)p->getMesgIDp());
		GRouter->sendPacket(resp, connections);
	}
	else
		MCC->addError("Backend::handleInit : mkInitRespPacket failed");

	delete resp;

	return 1;
}

void Backend::handleInitResp(GnutellaPacket *p)
{
	int		fc;
	int		fs;
	char		*body;

	if (!((body	= p->getBodyp())))
	{
		MCC->addError("handleInitResp : getBody failed!");
		return;
	}

	memcpy((char *)&fc, body + 6, 4);
	memcpy((char *)&fs, body + 10, 4);

#ifndef ROUTER
	hostsOnNet++;
	filesOnNet	+= fc;
	kbsOnNet	+= fs;
#endif
}

int Backend::handlePush(GnutellaPacket *p)
{
	int		fileID;
	int		hostip;
	short		hostport;
	char		*body;
	char		cid[CLIENTID_LENGTH];

	if (!((body	= p->getBodyp())))
	{
		MCC->addError("pushpacket without body");
		return 0;
	}

	memcpy(cid, body, CLIENTID_LENGTH);

	if (strncmp(cid, ourClientID, CLIENTID_LENGTH))
		return 1;

	fileID		= *((int *)(body + CLIENTID_LENGTH));
	hostip		= *((int *)(body + CLIENTID_LENGTH + 4));
	hostport	= *((short *)(body + CLIENTID_LENGTH + 8));

	fileServer->push(fileID, hostip, hostport);
	return	0;
}

void Backend::handleSearch(GnutellaPacket *p)
{
	Packet		*resp;
	char		body[MAX_SEARCH_BODY_SIZE];
	short		speed;
	Connection	*c;
	char		host[256];

	if (!((c = (Connection *)connections->findElement(p->getSourceConnection()))))
		return;
	
	if (!Inet::getHostName(host, 256))
		return;
	
	if (p->getBody((unsigned char *)body, (int)p->getDataLength()) == -1)
	{
		MCC->addError("handleSearch : getBody failed!!!");
		return;
	}
	
	speed	= *((short *)body);

#ifndef ROUTER
	bigBrother->addQuery(body + 2, speed);
#endif
	resp	= fileServer->mkSearchResp(body + 2);

	if (!resp)
		return;

	// port already set in FServer....
	resp->setMesgID((unsigned char *)p->getMesgIDp());
	resp->setTargetConnection(p->getSourceConnection());
	resp->setSearchRespIP(Inet::NSLookup(host));
	GRouter->sendPacket(resp, connections);
}

void Backend::handleNewConnection(Connection *c)
{
	int			error;
	GnutellaConnection	*gc;
//	Connection		*tc;
//	int			i;

	error	= 0;
	switch (c->getType())
	{
		case	CON_TYPE_UPLOAD	: fileServer->addUpload(c);
				break;
		case	CON_TYPE_DOWNLOAD	: error		= 1;
				break;
		case	CON_TYPE_GNUTELLA	:
				if (c->read() == -1)
				{
					error	= 1;
					break;
				}
				if (c->getInBufferBytes() < strlen("GNUTELLA CONNECT/0.4\n\n"))
				{
					error	= 1;
					break;
				}
				if (!c->skipBytes(strlen("GNUTELLA CONNECT/0.4\n\n")))
				{
					error	= 1;
					break;
				}
				/*
				 * Suggestion of Brian Macke : Allow max. 1 connection
				 * per IP addr. This prevent people from eating up all
				 * free slots in connections
				 */
#warning Disabled ip connection counter
/*
				slt	= sizeof(struct sockaddr);
				getsockname(c->getConnection(), (struct sockaddr *)&nsa, &slt);
				for (i = 0; i < connections->getSize(); ++i)
				{
					if (i == c->getEntry())
						continue;
					if (!((tc = (Connection *)connections->findElement(i))))
						continue;
					slt	= sizeof(struct sockaddr_in);
					getsockname(tc->getConnection(), (struct sockaddr *)&csa, &slt);
					if (nsa.sin_addr.s_addr == csa.sin_addr.s_addr)
					{
						error	= 1;
						break;
					}
				}
*/
				if (error == 1)
					break;

				c->sendStr("GNUTELLA OK\n\n", strlen("GNUTELLA OK\n\n"));
				gc	= new GnutellaConnection(c);
				delete c;

				if (connections->storeElement(gc) == ENTRY_INVALID)
				{
					error	= 1;
					break;
				}
				gc->setState(CON_STATE_ACTIVE);
				/*
				 * new policy : only send init when we are
				 * running low on hosts
				 */
				if (gc->isOutgoing() && (hosts->getElements() < 100))
				{
					Packet	*p;

					p	= new Packet(MCC->get_DEFAULT_TTL());
					p->mkInitPacket(MCC->get_DEFAULT_TTL());
					p->setMesgID((unsigned char *)ourInit->getMesgIDp());
					p->setTargetConnection(gc->getEntry());
					if (!GRouter->sendPacket(p, connections))
						MCC->addError("BUG : Backend : handleNewConnections"
								"insuffient buffersize");
					delete p;
				}
				if (gc->bufferSend() == -1)
					error	= 1;
				break;
		case	CON_TYPE_PUSH_OUT	: error         = 1;
				break;
		case	CON_TYPE_PUSH_IN	:
#ifndef ROUTER
					leechMan->handlePush(c);
#else
					error		= 1;
#endif
				break;
		default			: error		= 1;
				break;
	}

	if (error)
	{
		  c->close();
		  delete c;
	}
}

#ifndef ROUTER
int Backend::autoPush()
{
	int		i;
	DB		*downloads;
	Download	*d;

	if (!MCC->get_AUTO_PUSH())
		return 0;

	downloads	= leechMan->getDownloads();
	for (i = 0; i < downloads->getSize(); i++)
		if ((d = (Download *)downloads->findElement(i)))
		{
			if (d->getState() ==  CON_STATE_CON_FAILED)
			{
				Packet			*p;
				char			host[256];

				if (!Inet::getHostName(host, 256))
				{
					MCC->addError("gethostname failed!!!");
					downloads->deleteElement(i);
					continue;
				}

				p	= new Packet(MCC->get_DEFAULT_TTL());
				p->mkPushPacket(d->getCIDp(), d->getFileID(),
						Inet::NSLookup(host), MCC->get_PORT(),
						MCC->get_DEFAULT_TTL());
				p->setTargetConnection(d->getResultConnection());

				leechMan->newPushRequest(d->getFileID(), d->getFileNamep(),
							p->getMesgIDp(), d->getCIDp(), d->getOffset());

				GRouter->sendPacket(p, connections);
				killDownload(i);
				delete p;
			}
		}

	return 1;
}
#endif

int Backend::autoOpenConnections()
{
	Connection	*c;
	int		i;

	if (MCC->get_AUTO_CONNECTIONS() <= 0)
		return 0;

	for (i = 0; i < connections->getSize(); i++)
		if ((c = (Connection *)connections->findElement(i)))
		{
			switch (c->getState())
			{
				case CON_STATE_INACTIVE		:
				case CON_STATE_INIT		:
				case CON_STATE_CONNECTED	:
				case CON_STATE_LOGIN_SENT	:
				case CON_STATE_ACTIVE		:
								break;
				default	: closeConnection(i);
			}
 		}

 	if (	connections->isFull() ||
		(connections->getElements() >= MCC->get_MAX_CONNECTIONS()) ||
		(connections->getElements() >= MCC->get_AUTO_CONNECTIONS()) )
		return 0;

	while (hostIndex < hosts->getSize())
	{
		Host	*h;
		int	used;

		if (!((h = (Host *)hosts->findElement(hostIndex))))
		{
			hostIndex++;
			continue;
		}

		if (!h->isReal())
		{
			hostIndex++;
			continue;
		}

		used	= 0;
		for (i = 0; i < connections->getSize(); i++)
			if ((c = (Connection *)connections->findElement(i)))
				if (c->getHostIndex() == hostIndex)
				{
					used	= 1;
					break;
				}
		if (used)
		{
			hostIndex++;
			continue;
		}

		connectToHost(hostIndex++);
		if ( (connections->getElements() >= 
					MCC->get_AUTO_CONNECTIONS()) ||
				connections->isFull())
			break;
	}
	if (hostIndex == hosts->getSize())
		hostIndex	= 0;

	/*
	 * If there are less hosts available than there are
	 * connections wanted, there's no need to waste CPU time
	 * trying to connect.
	 */

	if (hosts->getElements() <= connections->getElements())
		return 0;

	return 1;
}

void Backend::newClientID(char *buf)
{
	unsigned int	seed;
	unsigned int	i;

	sprintf(buf, "%d", (int)time(0));
        seed    = buf[2] * buf[3];
        srand(seed);
        i       = rand();
	memcpy(buf + 4, (char *)&i, 4);
	i	= rand();
	memcpy(buf + 8, (char *)&i, 4);
        i       = rand();
        memcpy(buf + 12, (char *)&i, 4);
}

/*
 * GnutellaRouter Stats
 */
unsigned int Backend::getGRInitPackets()
{
	return GRouter->getInitPackets();
}

unsigned int Backend::getGRInitRespPackets()
{
	return GRouter->getInitRespPackets();
}

unsigned int Backend::getGRPushPackets()
{
	return GRouter->getPushPackets();
}

unsigned int Backend::getGRSearchPackets()
{
	return GRouter->getSearchPackets();
}

unsigned int Backend::getGRSearchRespPackets()
{
	return GRouter->getSearchRespPackets();
}

unsigned int Backend::getGRDoublePackets()
{
	return GRouter->getDoublePackets();
}

unsigned int Backend::getGRRespMatches()
{
	return GRouter->getRespMatches();
}

unsigned int Backend::getGRRespMisses()
{
	return GRouter->getRespMisses();
}

/*
 * HostCatcher Stats
 */
unsigned int Backend::getHCHostsCaught()
{
	return myHostCatcher->getHostsCaught();
}

/*
 * Networker Stats
 */
unsigned int Backend::getNWBytesIn()
{
	return netMan->getBytesIn();
}

unsigned int Backend::getNWBytesOut()
{
	return netMan->getBytesOut();
}
