/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include "Constants.h"
#include "Config.h"
#include "Socket.h"
#include "DBElement.h"
#include "Buffer.h"
#include "Connection.h"

Connection::Connection() : DBElement()
{
	hostIndex	= -1;
	type		= CON_TYPE_INVALID;
	state		= CON_STATE_INACTIVE;
	lock		= 0;
	outgoing	= 0;
	tries		= 0;
	timeStamp	= time(0);
	inBuffer	= new Buffer(CONNECTION_BUF_SIZE);
	outBuffer	= new Buffer(CONNECTION_BUF_SIZE);
	sock		= new Socket();
	hv		= 0;
}

Connection::Connection(int newSocket) : DBElement()
{
	sock		= new Socket(newSocket);
	hostIndex	= -1;
	type		= CON_TYPE_INVALID;
	state		= CON_STATE_INACTIVE;
	lock		= 0;
	outgoing	= 0;
	tries		= 0;
	timeStamp	= time(0);
	inBuffer	= new Buffer(CONNECTION_BUF_SIZE);
	outBuffer	= new Buffer(CONNECTION_BUF_SIZE);
	if (sock->getSock() >= 0)
		hv	= sock->getSock();
	else
		hv	= 0;
}

Connection::Connection(Socket *newSocket, int newHostIndex) : DBElement()
{
	sock		= newSocket;
	hostIndex	= newHostIndex;
	type		= CON_TYPE_INVALID;
	state		= CON_STATE_INACTIVE;
	lock		= 0;
	outgoing	= 0;
	tries		= 0;
	timeStamp	= time(0);
	inBuffer	= new Buffer(CONNECTION_BUF_SIZE);
	outBuffer	= new Buffer(CONNECTION_BUF_SIZE);
	if (sock->getSock() >= 0)
		hv	= sock->getSock();
	else
		hv	= 0;
}

Connection::Connection(Socket *newSocket) : DBElement()
{
	sock		= newSocket;
	hostIndex	= -1;
	type		= CON_TYPE_INVALID;
	state		= CON_STATE_INACTIVE;
	lock		= 0;
	outgoing	= 0;
	tries		= 0;
	timeStamp	= time(0);
	inBuffer	= new Buffer(CONNECTION_BUF_SIZE);
	outBuffer	= new Buffer(CONNECTION_BUF_SIZE);
	if (sock->getSock() >= 0)
		hv	= sock->getSock();
	else
		hv	= 0;
}

Connection::Connection(Connection *c) : DBElement()
{
	hostIndex	= c->getHostIndex();
	type		= c->getType();
	state		= c->getState();
	lock		= c->getLock();
	outgoing	= c->isOutgoing();
	tries		= c->getTries();
	timeStamp	= c->getTimeStamp();
	inBuffer	= c->copyInBuffer();
	outBuffer	= c->copyOutBuffer();
	sock		= new Socket(c->getConnection());
	sock->setIp(c->getIp());
	sock->setPort(c->getPort());
	if (sock->getSock() >= 0)
		hv	= sock->getSock();
	else
		hv	= 0;
}

Connection::~Connection()
{
	delete inBuffer;
	delete outBuffer;
	delete sock;
}

int Connection::isOutgoing()
{
	return outgoing;
}

void Connection::setOutgoing()
{
	outgoing	= 1;
}

int Connection::getState()
{
	return state;
}

void Connection::setState(int newState)
{
	state	= newState;
}

int Connection::getTries()
{
	return tries;
}

Buffer *Connection::copyInBuffer()
{
	return inBuffer->copyBuffer();
}

Buffer *Connection::copyOutBuffer()
{
	return outBuffer->copyBuffer();
}

int Connection::bufferSend()
{
	int	i;

	if (((i = outBuffer->bufSend(sock))) == -1)
		setState(CON_STATE_CON_ERROR);
	
	return i;
}

int Connection::sendStr(char *buf, unsigned int buflen)
{
	if (outBuffer->bytesEmpty() < buflen)
		return 0;

	return outBuffer->add(buf, buflen);
}

int Connection::read()
{
	return inBuffer->bufRecv(sock);
}

void Connection::setLock(int i)
{
	lock	= i;
}

int Connection::getLock()
{
	return lock;
}

int Connection::getInBuffer(char *buf, unsigned int range)
{
	return inBuffer->bufSuck(buf, range);
}

int Connection::peekInBuffer(char *buf, unsigned int range)
{
	return inBuffer->bufPeek(buf, range);
}

unsigned int Connection::getInBufferSize()
{
	return inBuffer->getSize();
}

unsigned int Connection::getOutBufferSize()
{
	return outBuffer->getSize();
}

unsigned int Connection::getInBufferBytes()
{
	return inBuffer->bytesInBuffer();
}

unsigned int Connection::getOutBufferBytes()
{
	return outBuffer->bytesInBuffer();
}

int Connection::getConnection()
{
	return sock->getSock();
}

int Connection::getType()
{
	return type;
}

int Connection::getHostIndex()
{
	return hostIndex;
}

int Connection::skipBytes(unsigned int bytes)
{
	return inBuffer->skip(bytes);
}

void Connection::setConnection(int newSocket)
{
	sock->setSocket(newSocket);
	if (sock->getSock() >= 0)
		hv	= sock->getSock();
	else
		hv	= 0;
}

void Connection::setHostIndex(int newHostIndex)
{
	hostIndex	= newHostIndex;
}

void Connection::close()
{
	sock->callClose();
	/*
	 * Do NOT adjust hv!!!!
	 */
}

void Connection::setType(int newType)
{
	type	= newType;
}

int Connection::tryToConnect()
{
	if (state > CON_STATE_ACTIVE)
		return -1;

	if (state == CON_STATE_INACTIVE)
	{
		switch (sock->callConnect(ip, port))
		{
			case SOCKET_CONNECT_SUCCESS :
						resetTimeStamp();
						setState(CON_STATE_CONNECTED);
						return 1;
			case SOCKET_CONNECT_IN_PROGRESS	:
						setState(CON_STATE_INIT);
						return 0;
			case SOCKET_CONNECT_FAILED :
						setState(CON_STATE_CON_FAILED);
						return -1;
			default			:
						fprintf(stderr, "ERROR : BUG : Connection::tryToConnect() : unknown return value "
								"from socket->callConnect()!\r\n");
						setState(CON_STATE_CON_FAILED);
						return -1;
		}
			
	}

	if (getAge() > CONNECTION_TIMEOUT)
	{
		setState(CON_STATE_CON_FAILED);
		return -1;
	}

	if (state != CON_STATE_INIT)
		return 0;

        switch (sock->testConnect())
	{
		case	SOCKET_CONNECT_FAILED		:
				this->setState(CON_STATE_CON_FAILED);
				break;
		case	SOCKET_CONNECT_IN_PROGRESS 	:
				break;
		case	SOCKET_CONNECT_SUCCESS	:
				resetTimeStamp();
				setState(CON_STATE_CONNECTED);
				return  1;
		default	: fprintf(stderr, "ERROR : BUG : Connection::tryToConnect() : testConnect is doing weird!\r\n");
				setState(CON_STATE_CON_ERROR);
				return -1;
	}

	return 0;
}

time_t Connection::getTimeStamp()
{
	return timeStamp;
}

void Connection::resetTimeStamp()
{
	timeStamp	= time(0);
}

time_t Connection::getAge()
{
	return	time(0) - timeStamp;
}

int Connection::socket()
{
	int	res;
	
	res	= sock->callSocket();
	if (sock->getSock() >= 0)
		hv	= sock->getSock();
	else
		hv	= 0;
	
	return res;
}

int Connection::Connect(unsigned long hostIp, unsigned short hostPort)
{
	state	= CON_STATE_INACTIVE;
	sock->callSocket();
	ip	= hostIp;
	port	= hostPort;
	setOutgoing();
	tryToConnect();

	return 1;
}

Socket *Connection::getSocket()
{
	return sock;
}

unsigned short Connection::getPort()
{
	return sock->getPort();
}
		
unsigned long int Connection::getIp()
{
	return sock->getIp();
}
		                
void Connection::getDottedIp(char *buf)
{
	sock->getDottedIp(buf);
}
		                
/*
 * DBElement Inheritance
 */

int Connection::equals(DBElement * con)
{
	return (sock->getSock() == ((Connection *)con)->getConnection());
}

int Connection::lessThan(DBElement * con)
{
	return (sock->getSock() < ((Connection *)con)->getConnection());
}

int Connection::greaterThan(DBElement * con)
{
	return (sock->getSock() > ((Connection *)con)->getConnection());
}

unsigned int Connection::hashValue()
{
	return (unsigned int)hv;
}

/*
 * FAME
 */
char *Connection::getInBufferp(unsigned int range)
{
	return inBuffer->bufSuckp(range);
}
