/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
 * Maximum number of times a connect is called before
 * we call it an error. Each try lasts 1 milisec.
 */
#define		MAX_CON_TRIES		3000 // 3 secs

/*
 * Possible states for a connection
 */
#define		CON_STATE_INACTIVE	0
#define		CON_STATE_INIT		1
#define		CON_STATE_CONNECTED	2
#define		CON_STATE_LOGIN_SENT	3
#define		CON_STATE_ACTIVE	4
#define		CON_STATE_DONE		5
#define		CON_STATE_CON_FAILED	6
#define		CON_STATE_LOGIN_FAILED	7
#define		CON_STATE_CON_ERROR	8
#define		CON_STATE_CANCEL	9
#define		CON_STATE_CLOSED	10
#define		CON_STATE_INVALID	99

/*
 * Connection types
 */
#define		CON_TYPE_UPLOAD		1
#define		CON_TYPE_DOWNLOAD	2
#define		CON_TYPE_GNUTELLA	3
#define		CON_TYPE_PUSH_OUT	4
#define		CON_TYPE_PUSH_IN	5
#define		CON_TYPE_INVALID	6

/*
 * How big can a gnutellapacket be ?
 * There are 2 packettypes with variable length : 
 * SEARCH and SEARCH_RESP
 * The length of a SEARCH PACKET is determined mostly by the length
 * of the string. Lets just assume a file name length is no longer
 * than 256 chars, plus 23 byte header = 279 bytes max.
 * a SEARCH RESP packet is bigger : 23 byte head, 10 bytes host-data
 * max 255 times a search rec and a 16 bytes clientID. A search Rec is 
 * at most 4 + 4 + 255 + 2 bytes = 265.
 * Thus, the max size = 23 + 10 + (255 * 265 ) + 16 = 67624 bytes
 * Never decrease the value of CONNECTION_BUF_SIZE to lower than that.
 */

#define CONNECTION_BUF_SIZE	68000		// Must be at least max size
#define CONNECTION_TIMEOUT	30

class Connection : public DBElement
{
	private:
		int			hv;
		
	protected:                      
		int			type;
 		int			hostIndex;
		int			lock;
		int			state;
		Buffer			*inBuffer;
		Buffer			*outBuffer;
	               int         		tries;
		int			outgoing;
		time_t			timeStamp;
		Socket			*sock;
		unsigned long int	ip;
		unsigned short		port;

	public	:
		Connection();

		/*
		 * newConnection is a socket
		 */
		Connection(int newSocket);

		Connection(Connection *c);

		Connection(Socket *s);

		Connection(int newConnection, unsigned long ip, unsigned short port);

		Connection(Socket *newSocket, int newHostIndex);

		~Connection();

		/*
		 * Warning ! this changes the hash value!
		 */
		int		socket();

		int		Connect(unsigned long ip, unsigned short port);

		int 		getConnection();

		int 		getHostIndex();

		/*
		 * newConnection is a socket
		 * Warning ! this changes the hash value!
		 */
		void 		setConnection(int newConnection);

		void 		setHostIndex(int newHostIndex);

		Buffer		*copyInBuffer();

		Buffer		*copyOutBuffer();

		/*
		 * Return Values for tryToConnect :
		 * 1	= connection established
		 * 0	= no connection yet
		 * -1	= MAX_TRIES Exceeded, connection failed
		 */
		int		tryToConnect();

		int		open();

		void		close();

		int		skipBytes(unsigned int bytes);

		/*
		 * returns bytes read
		 */
		int		read();

		int		sendStr(char *buf, unsigned int buflen);

		int		getLock();

		void		setLock(int i);

		int		getTries();

		/*
		 * Read from inBuffer. Returns bytes actually copied
		 */

		int		getInBuffer(char *buf, unsigned int range);

		int		peekInBuffer(char *buf, unsigned int range);

		/*
		 * return the size of the buffer
		 */
		unsigned int	getInBufferSize();

		unsigned int	getOutBufferSize();

		/*
		 * return the number of bytes in the buffer
		 */
		unsigned int	getInBufferBytes();

		unsigned int	getOutBufferBytes();

		/*
		 * return true on succes
		 */
		int		putToOutBuffer(char *buf, unsigned int buflen);

		/*
		 * return false on socket error
		 */
		int		bufferSend();

		/*
		 * Port and IP are in network byte order!
		 */
		unsigned short	getPort();
		
		unsigned long int	getIp();
		                
                void		getDottedIp(char *buf);
		                
                int         	getType();

		void		setType(int newType);

		int		getState();

		int		isOutgoing();

		void		setOutgoing();

		time_t		getTimeStamp();

		void		resetTimeStamp();

		time_t		getAge();

		void		setState(int newState);
		
		Socket		*getSocket();

		/*
		 * FAME
		 */
		char		*getInBufferp(unsigned int range);

		/*
		 * DBElement inheritence
		 */

		int		equals(DBElement * con);

		int 		lessThan(DBElement * con);

		int 		greaterThan(DBElement * con);

		unsigned int 	hashValue();
};
