/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include "DBElement.h"
#include "DB.h"

#define HASH(X)		(((X)->hashValue()) % hashSpace)

DB::DB(int size)
{
	elements	= 0;
	last 		= 0;
	oldest		= ENTRY_INVALID;
	youngest	= ENTRY_INVALID;
	DBSize		= size;
	hashSpace	= DEFAULT_HASH_SIZE;
	myDB		= (DBElement **)calloc(size + 1, sizeof(DBElement *));
	myDBHash	= (DBElement **)calloc(hashSpace, sizeof(DBElement *));
}

DB::DB(int size, int newHashSpace)
{
	elements	= 0;
	last 		= 0;
	oldest		= ENTRY_INVALID;
	youngest	= ENTRY_INVALID;
	DBSize		= size;

	if (newHashSpace < 1)
		hashSpace	= HASH_TINY;
	else if (newHashSpace > (5 * 1024 * 1024))
		hashSpace	= HASH_HUGE;
	else
		hashSpace	= newHashSpace;

	myDB		= (DBElement **)calloc(size + 1, sizeof(DBElement *));
	myDBHash	= (DBElement **)calloc(hashSpace, sizeof(DBElement *));
}

DB::~DB()
{
	delete [] myDB;
	delete [] myDBHash;
}

void DB::clearDB()
{
	int	i;

	for (i = 0; i < DBSize; i++)
		if (myDB[i])
		{
			delete(myDB[i]);
			myDB[i]	= 0;
		}
	for(i = 0; i < hashSpace; ++i)
		myDBHash[i]	= 0;

	elements	= 0;
	last 		= 0;
	oldest		= ENTRY_INVALID;
	youngest	= ENTRY_INVALID;
}

DBElement * DB::findElement(int index)
{
	if ( (index >= 0) && (index < DBSize) )
		return myDB[index];

	return 0;
}

int DB::containsElement(DBElement *el)
{
	DBElement	*dbElem;

	if (!el)
		return ENTRY_INVALID;

	dbElem = myDBHash[HASH(el)];
	while (dbElem)
	{
		if (el->equals(dbElem))
			return dbElem->getEntry();

		if (el->lessThan(dbElem))
			dbElem	= dbElem->getLeft();
		else
			dbElem	= dbElem->getRight();
	}

	return ENTRY_INVALID;
}

int DB::storeElement(DBElement *el)
{
	DBElement       *dbElem;
	int		insertHere;
	unsigned int	hash;

	if (!el)
		return ENTRY_INVALID;

	el->setLeft(0);
	el->setRight(0);

	/*
	 * Already there ?
	 */
	if (containsElement(el) != ENTRY_INVALID)
		return ENTRY_INVALID;

	/*
	 * Determine index and cleanup if nessicary
	 */

	if (isFull())
	{
		if ((insertHere = oldest) == ENTRY_INVALID)
		{
			fprintf(stderr, "ERROR : BUG : DB.cpp::storeElement() : "
					"Full but oldest = INVALID!!!\n");
			return ENTRY_INVALID;
		}
		insertHere	= oldest;

		if (myDB[oldest])
			deleteElement(oldest);
		else
			fprintf(stderr, "ERROR : BUG : DB.cpp::storeElement : "
					"Full but still room!\n");
	}
	else
	{
		if (!myDB[last])
		{
			insertHere	= last;
			if (++last == DBSize)
				last	= 0;
		}
		else
		{
			for (insertHere = 0; (myDB[insertHere] &&
					(insertHere < DBSize)); insertHere++)
				;
			if ((last = insertHere + 1) >= DBSize)
				last	= 0;
		}
	}

	if (insertHere	== DBSize)
	{
		fprintf(stderr, "ERROR : BUG : DB.cpp::storeElement : "
				"Not full and yet no place empty!!!\n");
		return ENTRY_INVALID;
	}

	/*
	 * Insert in array
	 */
	el->setEntry(insertHere);
	myDB[insertHere]	= el;

	/*
	 * do older / younger stuff
	 */
	if (oldest == ENTRY_INVALID)
		oldest		= insertHere;
	if (youngest == ENTRY_INVALID)
		youngest	= insertHere;
	else
	{
		myDB[youngest]->setYounger(insertHere);
		myDB[insertHere]->setOlder(youngest);
		youngest	= insertHere;
	}

	/*
	 * insert in tree
	 */
	hash	= HASH(el);
        dbElem	= myDBHash[hash];
	if (!dbElem)
	{
		myDBHash[hash]	= el;
		el->setParent(NO_PARENT);
	}
	else
	{
		while (1)
		{
			if (el->lessThan(dbElem))
			{
				if (dbElem->getLeft())
					dbElem	= dbElem->getLeft();
				else
				{
					dbElem->setLeft(el);
					break;
				}
			}
			else
			{
				if (dbElem->getRight())
					dbElem	= dbElem->getRight();
				else
				{
					dbElem->setRight(el);
					break;
				}
			}
		}
		el->setParent(dbElem->getEntry());
	}

	elements++;

	return insertHere;
}

int DB::removeElement(int index)
{
	DBElement	*leftchain;
	DBElement	*rightchain;
	DBElement	*el;
	DBElement	*tmp;
	int		l;

	if ( (index < 0) || (index >= DBSize) )
	{
		fprintf(stderr, "ERROR : BUG : DB.cpp::removeElement : "
				"Index out of range\n");
		return 0;
	}

	el		= myDB[index];
	if (!el)
	{
		fprintf(stderr, "ERROR : BUG : DB.cpp::removeElement : "
				"Element not found\n");
		return 0;
	}

	/*
	 * Do older / younger stuff
	 */
	if (youngest == index)
	{
		youngest	= el->getOlder();
		if (youngest != ENTRY_INVALID)
			myDB[youngest]->setYounger(ENTRY_INVALID);
	}
	else
		myDB[ el->getYounger() ]->setOlder(el->getOlder());

	if (oldest == index)
	{
		oldest		= el->getYounger();
		if (oldest != ENTRY_INVALID)
			myDB[oldest]->setOlder(ENTRY_INVALID);
	}
	else
		myDB[ el->getOlder() ]->setYounger(el->getYounger());


	/*
	 * removal from array. set last to point to this free entry
	 */
	myDB[index]	= 0;
	if ( myDB[last] || (index < last) )
		last	= index;

	/*
	 * remove from tree
	 */
	if ( (el->getParent() <= NO_PARENT) || (el->getParent() >= DBSize) )
	{
		myDBHash[HASH(el)]	= 0;
		if ( (el->getParent() < NO_PARENT) || (el->getParent() >= DBSize) )
			fprintf(stderr, "ERROR : BUG : DB.cpp::removeElement : "
					"Invalid parent (%d)in tree!!!\n", el->getParent());
 	}
	else
	{
		tmp	= myDB[el->getParent()];
		if (!tmp)
		{
			fprintf(stderr, "ERROR : BUG : DB.cpp::removeElement : "
					"Parent not found!!!\n");
			return 0;
		}
		if (tmp->getLeft() == el)
			tmp->setLeft(0);
		else
			if (tmp->getRight() == el)
				tmp->setRight(0);
			else
			{
				fprintf(stderr, "ERROR : BUG : DB.cpp::removeElement : "
						"Not Left Or Right.\n");
				return 0;
			}
	}

	/*
	 * Creating a linked list from the tree beneath el
	 */
	leftchain	= el->getLeft();
	rightchain	= el->getRight();
	leftchain	= this->mklist(leftchain);
	rightchain	= this->mklist(rightchain);
	if (leftchain)
	{
		tmp		= leftchain;
		for(l = 1; (tmp->getLeft()); l++)
                	tmp     = tmp->getLeft();
		if (rightchain)
		{
			tmp->setLeft(rightchain);
			l++;
		}
	}
	else
	{
		if ((leftchain = rightchain))
			l	= 1;
		else
			l	= 0;
	}

	// Count elements in rightchain
	if (rightchain)
		while ((rightchain 	= rightchain->getLeft()))
			l++;

	/*
	 * leftchain is now a sorted linked list, l is its length.
	 */

	if (leftchain)
		this->insertSortedList(leftchain, l, HASH(leftchain));

	elements--;

	return 1;
}

int DB::deleteElement(int index)
{
	DBElement	*el;

	if (!((el = myDB[index])))
		return 0;

	if (!removeElement(index))
		return 0;

	delete( el );
	return 1;
}

void DB::insertSortedList(DBElement *list, int len, int hash)
{
	DBElement	*tmp, *prv, *insertMe;
	int		i;

	if (len < 1)
		return;
	if (len == 1)
	{
		this->insertSortedElement(list, hash);
		return;
	}
	if (len == 2)
	{
		insertSortedElement(list->getLeft(), hash);
		insertSortedElement(list, hash);
		return;
	}

	prv	= list;
	tmp	= list->getLeft();
	for (i = 1; (i < (len / 2)); i++)
	{
		if (tmp)
		{
			prv	= tmp;
			tmp	= tmp->getLeft();
		}
	}
	prv->setLeft(0);
	insertMe	= tmp;
	tmp		= tmp->getLeft();
	insertSortedElement(insertMe, hash);
	insertSortedList(list, i, hash);
	insertSortedList(tmp, len - i - 1, hash);
}

void DB::insertSortedElement(DBElement *el, int hash)
{
	DBElement	*tmp;

	if (!el)
		return;

	el->setLeft(0);
	el->setRight(0);

	if (!(tmp	= myDBHash[hash]))
	{
		myDBHash[hash]	= el;
		el->setParent(NO_PARENT);
	}
	else
	{
		while(1)
		{
			if (el->equals(tmp))
				return;
/*
			{
				fprintf(stderr, "Error : DB "
						": insertSortedElement "
						": DOUBLE !!!\n");
				*(int *)0 = 0;
				return;
			}
*/			if (el->lessThan(tmp))
			{
				if (tmp->getLeft())
					tmp	= tmp->getLeft();
				else
				{
					tmp->setLeft(el);
					el->setParent(tmp->getEntry());
					break;
				}
			}
			else
			{
				if (tmp->getRight())
					tmp	= tmp->getRight();
				else
				{
					tmp->setRight(el);
					el->setParent(tmp->getEntry());
					break;
				}
			}
		}
	}
}

DBElement *DB::mklist(DBElement *tree)
{
	DBElement	*left;
	DBElement	*right;
	DBElement	*res;
	DBElement	*tmp;

	if (!tree)
		return 0;
	/*
	 * make a 'left' and a 'right' chain
	 */

	if (tree->getLeft())
		left	= mklist(tree->getLeft());
	else
		left	= 0;
	if (tree->getRight())
		right	= mklist(tree->getRight());
	else
		right	= 0;

	/*
	 * make res beginning point, add 'left' chain, then tree.
	 * tree should then be the last element so far.
	 */
	if (left)
	{
		res 	= tmp	= left;
		while (tmp->getLeft())
			tmp	= tmp->getLeft();
		tmp->setLeft(tree);
	}
	else
		res	= tree;

	/*
	 * Add 'right' chain
	 */
	tree->setLeft(right);

	return res;
}

int DB::isFull()
{
	return (elements == DBSize);
}

int DB::getSize()
{
	return DBSize;
}

int DB::getElements()
{
	return elements;
}

int DB::containsElementAtIndex(int i)
{
	if ( (i >= 0) && (i < DBSize) )
		return (int)myDB[i];

	return 0;
}

int DB::touch(int index)
{
	DBElement	*el;

	if ( (index < 0) || (index >= DBSize) )
	{
		fprintf(stderr, "ERROR : BUG : DB::touch() : index out of range!!!\n");
		return 0;
	}

	if (youngest == index)
		return 1;

	if (youngest == ENTRY_INVALID)
	{
		fprintf(stderr, "ERROR : BUG : DB::touch() : youngest == INVALID!!!\n");
		return 0;
	}

	if (!((el = myDB[index])))
	{
		fprintf(stderr, "ERROR : BUG : DB::touch() : index == NULL!!!\n");
		return 0;
	}

	myDB[ el->getYounger() ]->setOlder(el->getOlder());

	if (oldest == index)
	{
		oldest		= el->getYounger();
		if (oldest != ENTRY_INVALID)
			myDB[oldest]->setOlder(ENTRY_INVALID);
	}
	else
		myDB[ el->getOlder() ]->setYounger(el->getYounger());

	el->setOlder(youngest);
	myDB[ youngest ]->setYounger(index);
	youngest	= index;

	return 1;
}

int DB::integrity()
{
	int	i;
	int	t;

	for (i = 0; i < DBSize; i++)
		if (myDB[i])
			if (myDB[i]->getEntry() != i)
			{
				fprintf(stderr, "INTEGRITY 1 "
					"(%d != %d)\n",
					myDB[i]->getEntry(), i);
				*((int *)0) = 0;
				return 0;
			}

	t	= 0;
	for (i = 0; i < DBSize; i++)
		if (containsElementAtIndex(i))
			t++;

	if (t != elements)
	{
		fprintf(stderr, "INTEGRITY 2\n");
		*((int *)0) = 0;
		return 0;
	}

	if (!elements)
	{
		for (i = 0; i < DBSize; i++)
			if (myDB[i])
				if ( (myDB[i]->getLeft()) ||
						(myDB[i]->getRight()) )
				{
					fprintf(stderr, "INTEGRITY 3\n");
					*((int *)0) = 0;
					return 0;
				}

		for (i = 0; i < hashSpace; i++)
			if (myDBHash[i])
				{
					fprintf(stderr, "INTEGRITY 4\n");
					*((int *)0) = 0;
					return 0;
				}
	}

	return 1;
}

void DB::balance()
{
	int		i;

	for (i = 0; i < hashSpace; i++)
	{
		DBElement	*el, *chain, *tmp;
		int len;

		el		= myDBHash[i];
		if (!el)
			continue;

		myDBHash[i]	= 0;
		tmp	= chain	= mklist(el);
		for (len = 0; tmp; tmp = tmp->getLeft())
			len++;
		insertSortedList(chain, len, HASH(el));
	}
}

