#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "Socket.h"
#include "Constants.h"
#include "ControlCentre.h"
#include "DBElement.h"
#include "DB.h"
#include "Buffer.h"
#include "GnutellaPacket.h"
#include "Connection.h"
#include "GnutellaConnection.h"
#include "Host.h"
#include "Networker.h"

Networker::Networker(ControlCentre *newMCC)
{
	MCC			= newMCC;
	listening		= 0;
	pendingConnections	= new DB(MCC->get_MAX_CONNECTIONS() / 2, HASH_TINY);
	pendingPackets		= new DB(MCC->get_MAX_CONNECTIONS() + 1);
	bytesIn			= 0;
	bytesOut		= 0;
	sock			= new Socket();
}

Networker::~Networker()
{
	stopListening();
}

GnutellaPacket *Networker::waitTraffic(DB *connections)
{
	GnutellaPacket		*pstart;
	GnutellaPacket		*pprv;
	GnutellaPacket		*p;
	GnutellaConnection	*c;
	char			*buf;
	int			i;
	unsigned int		len;

	if (!connections->getElements())
		return 0;

	readAllConnections(connections);

	/*
	 * See what pending packets are ready
	 */

	pstart	= 0;
	for (i = 0; i < pendingPackets->getSize(); i++)
	{
		if (!(p = (GnutellaPacket *)pendingPackets->findElement(i)))
			continue;

		if (!(c = (GnutellaConnection *)connections->findElement(p->getSourceConnection())))
		{
			pendingPackets->deleteElement(i);
			continue;
		}

		if (c->getState() != CON_STATE_ACTIVE)
			continue;

		if (c->getInBufferBytes() >= p->getDataLength())
		{
			len	= p->getDataLength();
			if (!((buf = c->getInBufferp(len))))
			{
				MCC->addError("BUG : Networker : couldn't get pending packet body");
				continue;
			}

			pendingPackets->removeElement(i);
			if (p->setBodyDirect(buf, len) != len)
			{
				MCC->addError("BUG : Networker : couldn't set pending packet body");
				p->setType(FUNC_INVALID);
				pprv	= pstart;
				pstart	= p;
				pstart->setLeft(pprv);
				c->punchRecvd();
				c->setLock(0);
				continue;
			}
			pprv	= pstart;
			pstart	= p;
			pstart->setLeft(pprv);
			c->punchRecvd();
			c->setLock(0);
		}
	}

	/*
	 * Create new pending Packets. Any ready packets are added to the list.
	 */
	for (i = 0; i < connections->getSize(); i++)
	{
		if (!(c = (GnutellaConnection *)connections->findElement(i)))
			continue;

		if ((c->getLock()) || (c->getInBufferBytes() <= 0))
			continue;

		if (c->getInBufferBytes() >= GNUTELLA_HEADER_LENGTH)
			p	= new GnutellaPacket(c->getEntry(), MCC->get_DEFAULT_TTL());
		else
			continue;

		/*
		 * There is enough data to create a packet
		 */
		if (!((buf = c->getInBufferp(GNUTELLA_HEADER_LENGTH))))
		{
			/*
			 * This shouldn't be possible
			 */
			MCC->addError("BUG : Networker : Error reading from buffer (0),"
					" can't create packet!!");
 			delete p;
			continue;
		}
		p->readHeaderFromString((unsigned char *)buf);
		delete buf;

		if (p->getDataLength() > MAX_BODY_LENGTH)
		{
			/*
			 * Someone is pulling our leg / bandwidth
			 * This packet is not sane. There's no point in reading more from
			 * this connection, an oversized packet doesn't even fit in the buffer.
 			 */
			p->setBodyDirect(0, 0);
			p->setType(FUNC_INVALID);
			pprv	= pstart;
			pstart	= p;
			pstart->setLeft(pprv);
			c->punchRecvd();
			continue;
		}

		/*
		 * Nate Riffe discovered that on slow machines, data eventually
		 * clogs up on incoming connections. His proposed fix was to read
		 * all available packets from a connection, in stead of just the first.
		 * Let's see if that helps. In order to continue reading on a connection
		 * that delivered a complete packet, I added a i--; in Cases #1 and #2.
		 */

		/*
		 * Case #1 No body.
		 */
		if (!p->getDataLength())
		{
			pprv	= pstart;
			pstart	= p;
			pstart->setLeft(pprv);
			c->punchRecvd();
			i--;
			continue;
		}

		/*
		 * Case #2 Body is complete
		 */
		if (p->getDataLength() <= c->getInBufferBytes())
		{
			len	= p->getDataLength();
			if (!((buf	= c->getInBufferp(len))))
			{
				/*
				 * This shouldn't be possible
				 */
				MCC->addError("BUG : Networker : Error reading from buffer (2),"
						" can't create packet");
	 			delete p;
				continue;
			}
			if (p->setBodyDirect(buf, len) != len)
			{
				/*
				 * This shouldn't be possible
				 */
				MCC->addError("BUG : Networker : Can't set packet body");
				delete p;
				free(buf);
				continue;
			}
			pprv	= pstart;
			pstart	= p;
			pstart->setLeft(pprv);
			c->punchRecvd();
			i--;
			continue;
		}

		/*
		 * Case #3 Body is incomplete. lock connection
		 * and store pending packet.
		 */
		if (pendingPackets->isFull() ||
				(pendingPackets->storeElement(p) == -1) )
		{
			/*
			 * Now we're screwed!
			 * We've read half the packet, and can't leave it
			 * anywhere ! Solution : Read rest of packet, then
			 * drop it.
			 */
			MCC->addError("Can't store pending packet!!!");
			c->skipBytes(p->getDataLength());
			delete p;
		}
		else
			c->setLock(1);
	}
	return pstart;
}

int Networker::doSend(DB *connections)
{
	Connection	*c;
	int		i;
	int		bs;
	int		res;

	res	= 0;
	for (i = 0; i < connections->getSize(); ++i)
		if ((c = (Connection *)connections->findElement(i)))
			if (	(c->getState() >= CON_STATE_CONNECTED)	&&
				(c->getState() <= CON_STATE_ACTIVE)	&&
				(c->getType()  == CON_TYPE_GNUTELLA)	)
				if (((bs = c->bufferSend())) != -1)
				{
					res		= 1;
					bytesOut	+= bs;
				}

	return res;
}

void Networker::readAllConnections(DB *connections)
{
	int		poll;
	Connection	*c;
	int		br;

	/*
	 * Check events on connections
	 */

	for (poll = 0; poll < connections->getSize(); poll++)
		if ((c = (Connection *)connections->findElement(poll)))
		{
			if (c->getState() != CON_STATE_ACTIVE)
				continue;

			/*
			 * handle connections
			 */
			br = c->read();
			if (br == -1)
			{
//				c->close();
				c->close();
				connections->deleteElement(poll);
			}
			else
				bytesIn	+= br;
		}
}

Connection *Networker::waitConnection(DB *hosts)
{
	int			i;
	char			identifier[50];
	Connection		*c;
	Connection		*prv;
	Connection		*first;

	doListen(hosts);

	first		= 0;
	for (i = 0; i < pendingConnections->getSize(); i++)
		if ((c = (Connection *)pendingConnections->findElement(i)))
		{
			if (c->read() == -1)
			{
				c->close();
				pendingConnections->deleteElement(i);
				continue;
			}

			if (c->getInBufferBytes() < 10)
			{
				if (c->getAge()	> 30)
				{
					c->close();
					pendingConnections->deleteElement(i);
				}
				continue;
			}

			c->peekInBuffer(identifier, 49);
			identifier[49]	= 0;

			if (!strncmp(identifier, "GNUTELLA CONNECT/0.4\n\n",
					strlen("GNUTELLA CONNECT/0.4\n\n")))
			{
				if (c->getInBufferBytes() <
					strlen("GNUTELLA CONNECT/0.4\n\n"))
				{
					if (c->getAge()	> 30)
					{
						c->close();
						pendingConnections->deleteElement(i);
					}
					continue;
				}
				c->setType(CON_TYPE_GNUTELLA);
			}
			else if (!strncmp(identifier, "GET /get/",
					strlen("GET /get/")))
			{
				if (c->getInBufferBytes() < 12)
				{
					if (c->getAge()	> 30)
					{
						c->close();
						pendingConnections->deleteElement(i);
					}
					continue;
				}
				c->setType(CON_TYPE_UPLOAD);
			}
			else if (!strncmp(identifier, "GIV ",
					strlen("GIV ")))
			{
				if (c->getInBufferBytes() < 40)
				{
					if (c->getAge()	> 30)
					{
						c->close();
						pendingConnections->deleteElement(i);
					}
					continue;
				}
				c->setType(CON_TYPE_PUSH_IN);
			}
			else
			{
				c->close();
				pendingConnections->deleteElement(i);
				c	= 0;
			}

			if (c)
			{
				c->setState(CON_STATE_CONNECTED);
				pendingConnections->removeElement(i);
				prv	= first;
				first	= c;
				first->setLeft(prv);
			}
		}

	return first;
}

void Networker::doListen(DB *hosts)
{
	Connection		*c;
	Host			*h;
	Socket			*ns;
	int			HIndex;

	if ( (!isListening()) || hosts->isFull() ||
				pendingConnections->isFull() )
		return;
		
	if (!sock->testRecv(0))
		return;

	if (!((ns = sock->callAccept())))
		return;

	h	= new Host(ns->getIp(), ns->getPort());

	h->setReal(0);

	if (((HIndex = hosts->containsElement(h))) != ENTRY_INVALID)
		delete h;
	else if (hosts->isFull())
	{
		ns->callClose();
		delete h;
		return;
	}
	else if (((HIndex = hosts->storeElement(h))) == ENTRY_INVALID)
	{
		ns->callClose();
		delete h;
		return;
	}
	c	= new Connection(ns, HIndex);
	if (pendingConnections->storeElement(c) == ENTRY_INVALID)
	{
		c->close();
		delete c;
		return;
	}
}

int Networker::isListening()
{
	return listening;
}

int Networker::startListening()
{
	if (listening)
		return 0;

	if (!sock->callSocket())
	{
		MCC->addError("ERROR OPENING SOCKET!!!");
		return 0;
	}

	if (sock->callBind(MCC->get_PORT()) < 0)
	{
		MCC->addError("ERROR BINDING SOCKET!!!");
		sock->callClose();
		return 0;
	}

	sock->callListen(3);
	listening	= 1;

	return 1;
}

int Networker::stopListening()
{
	if (!listening)
		return 0;

	sock->callClose();
	listening	= 0;
	return 1;
}

unsigned int Networker::getBytesIn()
{
	return bytesIn;
}

unsigned int Networker::getBytesOut()
{
	return bytesOut;
}
