/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "DBElement.h"
#include "DB.h"
#include "RCRefList.h"
#include "RCSpam.h"
#include "RCPattern.h"
#include "RCMesgID.h"
#include "RCEntry.h"
#include "RCParameters.h"
#include "RCPacket.h"
#include "RCQPacket.h"
#include "RCQRPacket.h"
#include "RCPacketStorage.h"
#include "RCConnectionID.h"
#include "RCCache.h"

RCCache::RCCache(int newSize, int newMaxConnections)
{
	size			= newSize;
	lastFlush		= time(0);
	outList			= 0;
	connectionCounter	= 0;
	RCStore			= new RCPacketStorage(size);
	RCConnectionIDList	= 0;
	RCMesgIDMapDB		= 0;
	maxConnections		= newMaxConnections;
	
	startEntryInitTime	= DEFAULT_START_ENTRY_INIT_TIME;
	defaultSecondsToLive	= DEFAULT_DEFAULT_SECONDS_TO_LIVE;
	flushInterval		= DEFAULT_FLUSH_INTERVAL;
	
	QPacketsHandled		= 0;
	RPacketsHandled		= 0;
	packetsNotSent		= 0.0;
	packetsSent		= 0.0;
	hitCache		= 0;
	cacheError		= 0;
	cacheSpam		= 0;
	missedCache		= 0;
	hitCacheCon		= 0;
	missedCacheCon		= 0;
	hitZero			= 0;
	hitOK			= 0;
	hitSmallTTL		= 0;
	hitInitTime		= 0;
	respAdd			= 0;
	respKeptOld		= 0;
	respNotFound		= 0;
	respError		= 0;
}

RCCache::~RCCache()
{
	delete RCStore;
}

RCPacket *RCCache::onePacket()
{
	RCPacket	*p;
	
	if (outList)
	{
		p	= outList;
		outList	= outList->getNext();
		p->setNext(0);
		return p;
	}
	
	return 0;
}

int RCCache::mustFlush()
{
	return (time(0) > lastFlush + flushInterval);
}

double RCCache::getDoubleStat(int whatStat)
{
	switch (whatStat)
	{
		case	RC_STAT_PACKETS_NOT_SENT	: return packetsNotSent;
		case	RC_STAT_PACKETS_SENT		: return packetsSent;
	}
	fprintf(stderr, "ERROR : BUG : RCCache::getDoubleStat() : invalid"
					"stat requested %d\n", whatStat);
	return 0.0;
}

unsigned int RCCache::getStat(int whatStat)
{
	switch (whatStat)
	{
		case	RC_STAT_SIZE			: return size;
		case	RC_STAT_QPACKETS_HANDLED	: return QPacketsHandled;
		case	RC_STAT_RPACKETS_HANDLED	: return RPacketsHandled;
		
		case	RC_STAT_CACHE_ERROR		: return cacheError;
		case	RC_STAT_CACHE_SPAM		: return cacheSpam;
		case	RC_STAT_CACHE_HIT		: return hitCache;
		case	RC_STAT_CACHE_MISS		: return missedCache;
		
		case	RC_STAT_CACHE_HIT_CON		: return hitCacheCon;
		case	RC_STAT_CACHE_MISS_CON		: return missedCacheCon;

		case	RC_STAT_HIT_INIT_TIME		: return hitInitTime;
		case	RC_STAT_HIT_ZERO		: return hitZero;
		case	RC_STAT_HIT_SMALL_TTL		: return hitSmallTTL;
		case	RC_STAT_HIT_OK			: return hitOK;
	
		case	RC_STAT_RESP_ADD		: return respAdd;
		case	RC_STAT_RESP_KEPT_OLD		: return respKeptOld;
		case	RC_STAT_RESP_NOT_FOUND		: return respNotFound;
		case	RC_STAT_RESP_ERROR		: return respError;
	}
	fprintf(stderr, "ERROR : BUG : RCCache::getStat() : invalid stat requested %d\n", whatStat);
	return 0;
}

double RCCache::estimatePackets(int remainingTTL)
{
	/*
	 * assume every body has 4 connections,
	 * the packet would then be sent to 3 of them.
	 */
	double	r;
	double	p;
	 
	 r	= 0.0;
	 p	= 1.0;
	 
	 while (remainingTTL-- > 0)
	 {
//	 	p	*= (double)connectionCounter;
		p	*= 3;
	 	r	+= p;
	 }
	return r;
}

int RCCache::registerConnectionID(int cIndex)
{
	RCConnectionID	*rcc;
	
	if (connectionCounter == maxConnections)
	{
		fprintf(stderr, "ERROR : BUG : RCCache::registerConnectionID() : "
					"max connections reached!!\r\n");
		return 0;
	}
	
	for (rcc = RCConnectionIDList; rcc; rcc = rcc->getNext())
		if (cIndex == rcc->getConnectionID())
		{
			fprintf(stderr, "ERROR : BUG : RCCache::registerConnectionID() : "
						"double connection!!\r\n");
			return 0;
		}
	rcc			= new RCConnectionID(cIndex, RCConnectionIDList);
	RCConnectionIDList	= rcc;
	connectionCounter++;
		
	return 1;
}

int RCCache::isRegistered(int cIndex)
{
	RCConnectionID	*rcc;
	
	for (rcc = RCConnectionIDList; rcc; rcc = rcc->getNext())
		if (cIndex == rcc->getConnectionID())
			return 1;
		
	return 0;
}

int RCCache::unRegisterConnectionID(int cIndex)
{
	RCConnectionID	*rcc;
	RCConnectionID	*prv;
	
	prv	= 0;
	for (rcc = RCConnectionIDList; rcc; rcc = rcc->getNext())
	{
		if (cIndex == rcc->getConnectionID())
		{
			if (!prv)
				RCConnectionIDList	= rcc->getNext();
			else
				prv->setNext(rcc->getNext());
			
			connectionCounter--;
			delete rcc;
			return 1;
		}
		else
			prv	= rcc;
	}
		
	return 0;
}

int RCCache::addToOutList(RCPacket *p)
{
	p->setNext(outList);
	outList	= p;
	return 1;
}

/*
 * This function is the compare function for qsort
 * All it does is typecasting & dereferencing.
 */
int RCstrcmp(const void *arg1, const void *arg2)
{
	return (strcmp(*(char **)arg1, *(char **)arg2));
}

/*
 * This functions was written by Joris van Rantwijk because he couldn't stand
 * my own idea's about identifying cacheEntries based on a system of products
 * of prime numbers. He's right, this is better.
 */
void RCCache::sortPattern(char *pattern, int numStrings)
{
        char	*s;
	char	*p;
        int	i;
	char	**pList;

        s	= strdup(pattern);
        pList	= (char **)malloc(sizeof(char *) * numStrings);
        p	= s;

        for (i = 0; i < numStrings; i++)
	{
                p		= strtok(p, " ");
                pList[i]	= p;
		p		= 0;
	}

	/*
	 * Given the fact that sorting is a time-consuming
	 * business, I decided to do a simple swap when there are
	 * only 2 strings involved
	 */
	if (numStrings == 2)
	{
		char	*tmp;

		if (strcmp(pList[0], pList[1]) > 0)
		{
			tmp		= pList[0];
			pList[0]	= pList[1];
			pList[1]	= tmp;
		}
	}
	else
		qsort(pList, numStrings, sizeof(char *), RCstrcmp);
//	        quickSortPattern(pList, numStrings);

	p	= (char *)pattern;;
	*p	= 0;
	for (i = 0; i < numStrings; i++)
	{
		strcat(p, pList[i]);
	        p	+= strlen(p);
	        *p++	= ' ';
		*p	= 0;
	}
	*(p - 1)	= 0;

        free(s);
        free(pList);
}

#define RCPATTERN_VALIDCHARS	"*?ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890-+_=().\\'`\""

int RCCache::preparePattern(char *buf)
{
	int	i;
	int	j;
	int	res;

	res	= 0;
	j	= 0;
	for (i = 0; buf[i] && (i < MAX_QUERY_SIZE); ++i)
	{
		buf[i + j]	= toupper(buf[i]);
		if (strchr(RCPATTERN_VALIDCHARS, buf[i + j]))
		{
			if (	((i + j) > 0)	&& (buf[i + j - 1] == ' ') ||
				(!i)	)
				res++;
		}
		else
		{
			buf[i + j]	= ' ';
			if (	( (i > 0) && (buf[i + j - 1] == ' ') ) 	||
				( (!i) 	&& (buf[i + j] == ' '))		||
				(buf[i] == '\'') || (buf[i] == '"')	)
				j--;
		}
	}
	if (buf[i + j - 1] == ' ')
		buf[i + j - 1]	= 0;
	else
		buf[i + j]	= 0;

	return res;
}

