/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
 * Parameters
 */
#define DEFAULT_START_ENTRY_INIT_TIME		 30
#define DEFAULT_DEFAULT_SECONDS_TO_LIVE		900
#define DEFAULT_FLUSH_INTERVAL	 	 	 60

/*
 * Return values for handleQPacket()
 */
#define QUERY_SPAM				-2
#define QUERY_ERROR				-1
#define QUERY_CACHE_MISS 			 0
#define QUERY_CACHE_HIT		 		 1

/*
 * Return values for handleQRPacket()
 */
#define QUERY_RESPONSE_ADDED			 1
#define QUERY_RESPONSE_REPLACED_OLD		 2
#define QUERY_RESPONSE_MESGID_NOT_FOUND		-1
#define QUERY_RESPONSE_ERROR			-2
#define QUERY_RESPONSE_KEPT_OLD		 	-3

/*
 * Statistics that can be returned by getStats()
 */
#define RC_STAT_SIZE				10
#define RC_STAT_QPACKETS_HANDLED		11
#define RC_STAT_RPACKETS_HANDLED		12
#define RC_STAT_PACKETS_NOT_SENT		13
#define RC_STAT_PACKETS_SENT			14

#define RC_STAT_CACHE_HIT			20
#define RC_STAT_CACHE_MISS			21
#define RC_STAT_CACHE_ERROR			22
#define RC_STAT_CACHE_SPAM			23

#define RC_STAT_CACHE_HIT_CON			30
#define RC_STAT_CACHE_MISS_CON			31

#define RC_STAT_HIT_INIT_TIME			40
#define RC_STAT_HIT_ZERO			41
#define RC_STAT_HIT_SMALL_TTL			42
#define RC_STAT_HIT_OK				43

#define RC_STAT_RESP_ADD			50
#define RC_STAT_RESP_KEPT_OLD			51
#define RC_STAT_RESP_NOT_FOUND			52
#define RC_STAT_RESP_ERROR			53

class RCCache
{
	protected:
		RCPacket	*outList;
		RCPacketStorage	*RCStore;
		RCConnectionID	*RCConnectionIDList;
		DB		*RCMesgIDMapDB;
		time_t		lastFlush;
		int		size;
		int		connectionCounter;
		int		maxConnections;

		/*
		 * Tuneable stuff
		 */
		int		startEntryInitTime;
		int		defaultSecondsToLive;
		int		flushInterval;
		
		/*
		 * stats
		 */
		unsigned int	QPacketsHandled;
		unsigned int	RPacketsHandled;
		double		packetsNotSent;
		double		packetsSent;
		unsigned int	cacheError;
		unsigned int	cacheSpam;
		unsigned int	hitCache;
		unsigned int	missedCache;
		unsigned int	hitCacheCon;
		unsigned int	missedCacheCon;
		unsigned int	hitZero;
		unsigned int	hitOK;
		unsigned int	hitSmallTTL;
		unsigned int	hitInitTime;
		unsigned int	respAdd;
		unsigned int	respNotFound;
		unsigned int	respKeptOld;
		unsigned int	respError;
		 
		 /*
		  * local functions
		  */
		double		estimatePackets(int remainingTTL);
		int		addToOutList(RCPacket *p);
		void		sortPattern(char *pattern, int numStrings);
		int		preparePattern(char *buf);		


	public:
		RCCache(int newSize, int newMaxConnections);
		
		virtual ~RCCache() = 0;

		/*
		 * Returns either QUERY_RESPONSE_ADDED, QUERY_RESPONSE_REPLACED_OLD,
		 * QUERY_RESPONSE_MESGID_NOT_FOUND, QUERY_RESPONSE_ERROR or
		 * QUERY_RESPONSE_KEPT_OLD.
		 * Remember to pass the packet on.
		 */
		virtual int	handleResponsePacket(RCPacket *p) = 0;

		/*
		 * Returns either QUERY_SPAM, 
		 * QUERY_CACHE_MISS or
		 * QUERY_CACHE_HIT_PASS_ON
		 * After a call to this method the packets can be fetched with a call to outList().
		 */
		virtual int	handleQueryPacket(RCPacket *p) = 0;

		/*
		 * Returns a packet that must be sent or 0. Several calls may be nessicary 
		 * before there are no more packets. All packets will have their 
		 * targetConenctionID set correctly, please note the value 'CONNECTIONID_BROADCAST'
		 * which indicates that the packet should be send to all connections except the
		 * sourceConnection of the packet.
		 */
		RCPacket	*onePacket();

		/*
		 * returns true when it is time to flush
		 */
		int		mustFlush();
		
		/*
		 * Returns true when a flush was done successfully.
		 * Flush also deletes packets in the cache.
		 */
		virtual int	flush() = 0;

		unsigned int	getStat(int whatStat);
		
		double		getDoubleStat(int whatStat);
		
		/*
		 * You must register and unregister all connections here. Otherwise it cannot be
		 * determined which packets should be sent where.
		 */
		int		registerConnectionID(int cIndex);
		
		int		unRegisterConnectionID(int cIndex);

		int		isRegistered(int cIndex);
};
