/*****************************************************************************
 *  									*
 *   This program is free software; you can redistribute it and/or modify  		*
 *   it under the terms of the GNU General Public License as published by  		*
 *   the Free Software Foundation; either version 2 of the License, or     		*
 *   (at your option) any later version.                                   			*
 *                                                                         				*
 ***************************************************************************/

#include <stdio.h>
#include <time.h>
#include "DBElement.h"
#include "DB.h"
#include "RCRefList.h"
#include "RCSpam.h"
#include "RCPattern.h"
#include "RCMesgID.h"
#include "RCEntry.h"
#include "RCParameters.h"
#include "RCPacket.h"
#include "RCQPacket.h"
#include "RCQRPacket.h"
#include "RCPacketStorage.h"
#include "RCConnectionID.h"
#include "RCCache.h"
#include "RCQCache.h"

#ifndef CLIENTID_LENGTH
#define CLIENTID_LENGTH		16
#endif

#define RCMESGID_IS_MAPPED		0
#define RCMESGID_IS_ORIGINAL		1
#define RCMESGID_MUST_MAP		2
#define RCMESGID_IS_UNKNOWN	3

#define BUG(X)	fprintf(stderr, "ERROR : BUG : RCQCache::"X"\r\n")

RCQCache::RCQCache(int newSize, int newMaxConnections) : RCCache(newSize, newMaxConnections)
{
	RCPatternDB		= new DB(size, HASH_MEDIUM);
	RCMesgIDDB		= new DB(maxConnections * size, HASH_MEDIUM);
	RCParametersDB		= new DB(maxConnections * size, HASH_MEDIUM);
	RCEntryDB		= new DB(maxConnections * size, HASH_MEDIUM);
	RCSpamDB		= new DB(maxConnections * size, HASH_MEDIUM);
}

RCQCache::~RCQCache()
{
	RCPatternDB->clearDB();
	RCMesgIDDB->clearDB();
	RCParametersDB->clearDB();
	RCEntryDB->clearDB();
	RCSpamDB->clearDB();
	RCMesgIDMapDB->clearDB();

	delete RCPatternDB;
	delete RCMesgIDDB;
	delete RCParametersDB;
	delete RCEntryDB;
	delete RCSpamDB;
}

int RCQCache::handleQueryPacket(RCPacket *p)
{
	RCPattern	*pat;
	RCMesgID	*rcmid;
	RCConnectionID	*rcd;
	int		subStrs;
	char		pattern[MAX_QUERY_SIZE];
	char		mid[RCMESGID_LENGTH];
	int		pIndex;

	QPacketsHandled++;

	p->getMesgID(mid);
	rcmid	= new RCMesgID(0, mid);
	if (RCMesgIDDB->containsElement(rcmid) != ENTRY_INVALID)
	{
		delete rcmid;
		BUG("handeQueryPacket() : Already seen this request!");
		cacheError++;
		return QUERY_ERROR;
	}
	delete rcmid;

	((RCQPacket *)p)->getPattern(pattern);
	pattern[MAX_QUERY_SIZE - 1]	= 0;
	if ((subStrs = this->preparePattern(pattern)) < 1)
	{
		cacheError++;
		return QUERY_ERROR;
	}

	this->sortPattern(pattern, subStrs);
	pat	= new RCPattern(pattern);

	if ((pIndex = RCPatternDB->containsElement(pat)) == ENTRY_INVALID)
	{
		RCQPacket	*np;

       		/*
       		 * Update stats.
       		 */
       		missedCache++;

		 for (rcd = RCConnectionIDList; rcd; rcd = rcd->getNext())
       			if (rcd->getConnectionID() != p->getSourceConnectionID())
       			{
				packetsSent	+= estimatePackets(((RCQPacket *)p)->getTTL());
				np	= (RCQPacket *)p->duplicate();
				np->setTargetConnectionID(rcd->getConnectionID());
				this->addToOutList(np);
			}

		this->addQuery((RCQPacket *)p, pat);

		return QUERY_CACHE_MISS;
	}

	delete pat;

	pat	= (RCPattern *)RCPatternDB->findElement(pIndex);
	hitCache++;

	if (this->checkForSpam(pIndex, p->getSourceConnectionID()))
	{
		cacheSpam++;
		return QUERY_SPAM;
	}

	/*
	 * This is the real stuff.
        	 */

	rcmid	= 0;
	for (rcd = RCConnectionIDList; rcd; rcd = rcd->getNext())
	{
        		RCParameters	*rcpar;
        		RCPacket	*duplicate;
        		int		conID;
        		int		parIndex;

		/*
 		 * Skip source connection.
 		 */
		conID	= rcd->getConnectionID();
		if (conID == p->getSourceConnectionID())
			continue;

		/*
		 * Find parameters for this connection.
		 */
		rcpar	= new RCParameters(pIndex, conID, 0, 0);

		if ((parIndex = RCParametersDB->containsElement(rcpar)) == ENTRY_INVALID)
		{
		        	handleCacheMissOnConnection((RCQPacket *)p, rcpar, pat, conID);
		        	continue;
		}
		delete rcpar;

		rcpar	= (RCParameters *)RCParametersDB->findElement(parIndex);

		hitCacheCon++;

		/*
		 * If our current search TTL does not cover the TTL of the packet,
		 * pass it on.
		 */
		if (((RCQPacket *)p)->getTTL() > rcpar->getSearchTTL())
		{
			/*
			 * Pass on, refresh parameters.
			 */
			duplicate	= p->duplicate();
			duplicate->setTargetConnectionID(conID);
			this->addToOutList(duplicate);
			rcpar->setSearchTTL(((RCQPacket *)p)->getTTL());
			rcpar->setInitTime(time(0) + startEntryInitTime);

			packetsSent	+= estimatePackets(((RCQPacket *)p)->getTTL());
			hitSmallTTL++;
       			continue;
       		}

        	/*
	         * Our entries cover enough of the network.
        	 * If the initialisation time has passed respond
	         */
        		if (time(0) > rcpar->getInitTime())
		{
			this->handleCacheHitOnConnection((RCQPacket *)p, pat, pIndex, conID);
			continue;
		}

		/*
		 * We hit during initialisation time.
		 */
		if (!rcmid)
		{
			rcmid	= new RCMesgID(pIndex, mid);
			handleCacheHitInitTimeOnConnection((RCQPacket *)p, rcmid, pat);
		}
	}

	return QUERY_CACHE_HIT;
}

int RCQCache::handleResponsePacket(RCPacket *p)
{
	RCMesgID	*mesgID;
	RCEntry		*e;
	RCParameters	*par;
	RCQRPacket	*duplicate;
	char		mid[RCMESGID_LENGTH];
	int		mIdx;
	int		pIdx;
	int		eIdx;

	RPacketsHandled++;

	/*
	 * Use MesgId to find patternID
	 */
	p->getMesgID(mid);
	mesgID	= new RCMesgID(0, mid);
	if ((mIdx = RCMesgIDDB->containsElement(mesgID)) == ENTRY_INVALID)
	{
		/*
		 * This happens when a search contained an invalid
		 * pattern causing the query to be ignored.
		 */
//		BUG("handleResponsePacket() : MesgID not found!");
		delete mesgID;
		return QUERY_RESPONSE_MESGID_NOT_FOUND;
	}
	delete mesgID;

	mesgID	= (RCMesgID *)RCMesgIDDB->findElement(mIdx);
	pIdx	= mesgID->getPatternID();

	/*
	 * Sanity check : make sure we have the parameters
	 */
	par	= new RCParameters(pIdx, p->getSourceConnectionID(), 0 , 0);
	if (RCParametersDB->containsElement(par) == ENTRY_INVALID)
	{
		respError++;
		delete par;
		fprintf(stderr, "handleResponsePacket() : parameters not found for %d, %d!",
				pIdx, p->getSourceConnectionID());
		return QUERY_RESPONSE_ERROR;
	}
	delete	par;

	e	= new RCEntry(pIdx, p->getSourceConnectionID());
	if ((eIdx = RCEntryDB->containsElement(e)) == ENTRY_INVALID)
	{
		respError++;
		delete e;
		BUG("handleResponsePacket() : entry not found!");
		return QUERY_RESPONSE_ERROR;
	}
	delete e;

	/*
	 * Try to store the packet.
	 */
	e	= (RCEntry *)RCEntryDB->findElement(eIdx);
	if (e->getPacketID() == PKTID_INVALID)
		e->setPacketID(RCStore->findFreeIndex());
	if (e->getPacketID() == PKTID_INVALID)
	{
		BUG("handleResponsePacket() : No PacketID available!");
		return QUERY_RESPONSE_ERROR;
	}

//	if (RCStore->integrityCheck() < 0) *(int *)0=0;
	duplicate	= (RCQRPacket *)p->duplicate();
	duplicate->setPacketID(e->getPacketID());
	if (!RCStore->storePacket(duplicate))
	{
		respKeptOld++;
		delete duplicate;
		return QUERY_RESPONSE_KEPT_OLD;
	}
//	if (RCStore->integrityCheck() < 0) *(int *)0=0;

	respAdd++;
	e->incCounter();

	return QUERY_RESPONSE_ADDED;
}

int RCQCache::removeByPatternID(int pIndex)
{
	RCPattern	*p;
	RCMesgID	*rcmid;
	RCEntry		*rce;
	RCRefList	*rl;

//	if (RCStore->integrityCheck() < 0) *(int *)0=0;
//	this->checkIntegrity();
	p	= (RCPattern *)RCPatternDB->findElement(pIndex);

	for (rl = p->getRefList(RCPATTERN_REFLIST_SPAM); rl; rl = rl->getNext())
		RCSpamDB->deleteElement(rl->getIndex());

	for (rl = p->getRefList(RCPATTERN_REFLIST_PARAMETERS); rl; rl = rl->getNext())
		RCParametersDB->deleteElement(rl->getIndex());

	for (rl = p->getRefList(RCPATTERN_REFLIST_MESGID); rl; rl = rl->getNext())
		if ((rcmid = (RCMesgID *)RCMesgIDDB->findElement(rl->getIndex())))
			RCMesgIDDB->deleteElement(rl->getIndex());

	for (rl = p->getRefList(RCPATTERN_REFLIST_ENTRY); rl; rl = rl->getNext())
		if ((rce = (RCEntry *)RCEntryDB->findElement(rl->getIndex())))
		{
			if (rce->getCounter() != COUNTER_ZERO)
				RCStore->deleteByPacketID(rce->getPacketID());
			RCEntryDB->deleteElement(rl->getIndex());
		}

	RCPatternDB->deleteElement(pIndex);
//	this->checkIntegrity();
//	if (RCStore->integrityCheck() < 0) *(int *)0=0;

	return 1;
}

int RCQCache::addQuery(RCQPacket *p, RCPattern *pat)
{
	RCEntry		*rce;
	RCParameters	*rcpar;
	RCSpam		*rcs;
	RCMesgID	*rcmid;
	RCConnectionID	*rcc;
	RCRefList	*rl;
	int		pIndex;
	int		index;
	char		mid[RCMESGID_LENGTH];

	if (RCPatternDB->isFull())
	{
//		this->checkIntegrity();
	 	removeByPatternID(RCPatternDB->getOldest());
	}

	if ((pIndex = RCPatternDB->storeElement(pat)) == ENTRY_INVALID)
	{
		BUG("addQuery : store pattern failed!");
		delete pat;
		return 0;
	}

	/*
	 * new entry for all other connections
	 */
	for (rcc = RCConnectionIDList; rcc; rcc = rcc->getNext())
		if (rcc->getConnectionID() != p->getSourceConnectionID())
		{
	        		rce	= new RCEntry(pIndex, rcc->getConnectionID());
		        	if ((index = RCEntryDB->storeElement(rce)) == ENTRY_INVALID)
	        		{
	        			delete rce;
	        			this->removeByPatternID(pIndex);
				BUG("addQuery() : Adding Entry failed!");
	        			return 0;
	        		}
		        	pat->incRefCount();
		        	rl	= new RCRefList(index);
	        		if (!pat->addToRefList(RCPATTERN_REFLIST_ENTRY, rl))
		       	{
       				BUG("addQuery() : add reference to entry failed!");
		        		delete rl;
		        	}
		}

	/*
	 * new mesgid
	 */
	p->getMesgID(mid);
	rcmid	= new RCMesgID(pIndex, mid);
	if ((index = RCMesgIDDB->storeElement(rcmid)) == ENTRY_INVALID)
	{
		delete rcmid;
       		this->removeByPatternID(pIndex);
		BUG("addQuery() : Adding mesgID failed!");
       		return 0;
	}
       	rl	= new RCRefList(index);
       	if (!pat->addToRefList(RCPATTERN_REFLIST_MESGID, rl))
       	{
       		BUG("addQuery() : add reference to mesgID failed!");
       		delete rl;
       	}

	/*
	 * new spam
	 */
	 rcs	= new RCSpam(pIndex, p->getSourceConnectionID());
	 if ((index = RCSpamDB->storeElement(rcs)) == ENTRY_INVALID)
	 {
	 	delete rcs;
       		this->removeByPatternID(pIndex);
		BUG("addQuery() : Adding spam failed!");
       		return 0;
	 }
       	rl	= new RCRefList(index);
       	if (!pat->addToRefList(RCPATTERN_REFLIST_SPAM, rl))
       	{
       		BUG("addQuery() : add reference to spam failed!");
       		delete rl;
       	}

	/*
	 * new parameters
	 */
	for (rcc = RCConnectionIDList; rcc; rcc = rcc->getNext())
		if (rcc->getConnectionID() != p->getSourceConnectionID())
		{
	        		rcpar	= new RCParameters(pIndex, rcc->getConnectionID(),
	        					p->getTTL(), time(0) + startEntryInitTime);

		        	if ((index = RCParametersDB->storeElement(rcpar)) == ENTRY_INVALID)
		        	{
	        			delete rcpar;
	        			this->removeByPatternID(pIndex);
				BUG("addQuery() : Adding parameters failed!");
		        		return 0;
		        	}
		       	rl	= new RCRefList(index);
		       	if (!pat->addToRefList(RCPATTERN_REFLIST_PARAMETERS, rl))
		       	{
       				BUG("addQuery() : add reference to parameters failed!");
       				delete rl;
       			}
	        	}

	return 1;
}

int RCQCache::unRegisterConnectionID(int cIndex)
{
	RCEntry		*rce;
	RCParameters	*rcpar;
	RCSpam		*rcs;
	int		pIndex;

	if (!RCCache::unRegisterConnectionID(cIndex))
	{
		BUG("unRegisterConnectionID() : can't unregister in super!!");
		return 0;
	}

//	this->checkIntegrity();

	for (pIndex = 0; pIndex < RCPatternDB->getSize(); pIndex++)
	{
		RCPattern	*p;
		int		index;

		if (!((p = (RCPattern *)RCPatternDB->findElement(pIndex))))
			continue;

		rce	= new RCEntry(pIndex, cIndex);
		if ((index = RCEntryDB->containsElement(rce)) == ENTRY_INVALID)
			delete rce;
		else
		{
			delete rce;
			rce	= (RCEntry *)RCEntryDB->findElement(index);

			if (p->decRefCount() <= 0)
			{
//				if (RCStore->integrityCheck() < 0) *(int *)0=0;
				removeByPatternID(pIndex);
//				if (RCStore->integrityCheck() < 0) *(int *)0=0;
			}
			else
			{
//				if (RCStore->integrityCheck() < 0) *(int *)0=0;
				if (rce->getCounter() != COUNTER_ZERO)
					RCStore->deleteByPacketID(rce->getPacketID());
//				this->checkIntegrity();
//				if (RCStore->integrityCheck() < 0) *(int *)0=0;
				RCEntryDB->deleteElement(index);
				p->deleteFromRefList(RCPATTERN_REFLIST_ENTRY, index);
			}
		}

//		this->checkIntegrity();

		rcpar	= new RCParameters(pIndex, cIndex, 0 , 0);
		if ((index = RCParametersDB->containsElement(rcpar)) == ENTRY_INVALID)
			delete rcpar;
		else
		{
			delete rcpar;
			RCParametersDB->deleteElement(index);
			p->deleteFromRefList(RCPATTERN_REFLIST_PARAMETERS, index);
		}

		rcs	= new RCSpam(pIndex, cIndex);
		if ((index = RCEntryDB->containsElement(rcs)) == ENTRY_INVALID)
			delete rcs;
		else
		{
			delete rcs;
			RCSpamDB->deleteElement(index);
			p->deleteFromRefList(RCPATTERN_REFLIST_SPAM, index);
		}
//		this->checkIntegrity();
	}
//	this->checkIntegrity();

	return 1;
}

int RCQCache::flush()
{
	int		i;
	time_t		t;
	int		pktID;
	RCQRPacket	*p;

//	if (RCStore->integrityCheck() < 0) *(int *)0=0;
//	this->checkIntegrity();
	t	= time(0);
	for (i = 0; i < RCEntryDB->getSize(); i++)
	{
		RCEntry	*e;

		if (!((e = (RCEntry *)RCEntryDB->findElement(i))))
			continue;

		if ((pktID = e->getCounter()) == COUNTER_ZERO)
			continue;
		p	= (RCQRPacket *)RCStore->packetList(pktID);
//		RCStore->integrityCheckByPacketID(pktID);
		while (p)
		{
			RCQRPacket	*current;

			current	= p;
			p	= (RCQRPacket *)p->getNext();

			if (t + flushInterval < current->getFlushTime())
				continue;

			if (e->decCounter() < 0)
			 	BUG("flush() : counter smaller than zero!");

			if (current->getPacketID() != pktID)
			{
			 	BUG("flush() : counter smaller than zero!");
				continue;
			}

//			if (RCStore->integrityCheckByPacketID(pktID) < 0) *(int *)0=0;
			RCStore->deletePacket(current);
//			if (RCStore->integrityCheckByPacketID(pktID) < 0) *(int *)0=0;
		}
//		RCStore->integrityCheckByPacketID(pktID);

		/*
		 * Clean up when entry is obsolete....
		 */
		if ( 	(e->getCounter() == 0) 	||
			(t - e->getBorn() > defaultSecondsToLive) )
		{
			RCPattern	*pat;
			int		patternID;

//			if (RCStore->integrityCheck() < 0) *(int *)0=0;
			patternID	= e->getPatternID();
		 	RCEntryDB->deleteElement(pktID);
			if ((pat = (RCPattern *)RCPatternDB->findElement(patternID)))
			{
				if (pat->decRefCount() > 0)
					continue;
				 /*
				  * Pattern refcount hit zero. More cleaning to do.
				  * First remove pattern from patternDB,
				  * then remove all references to it from
				  * the other DB's.
				  */
				this->removeByPatternID(patternID);
			}
			else
				BUG("flush() : pattern not found! !!!");
//			if (RCStore->integrityCheck() < 0) *(int *)0=0;
		}
	}

//	if (RCStore->integrityCheck() < 0) *(int *)0=0;
	lastFlush	= t;

//	this->checkIntegrity();
	return 1;
}

int RCQCache::checkForSpam(int pIndex, int sConID)
{
	int	sIndex;
	RCSpam	*rcs;

	rcs	= new RCSpam(pIndex, sConID);

	if ((sIndex = RCSpamDB->containsElement(rcs)) != ENTRY_INVALID)
	{
        		delete rcs;

        		if ((rcs = (RCSpam *)RCSpamDB->findElement(sIndex)))
        		{
        			if (rcs->incCounter() > MAX_QUERY_BEFORE_SPAM)
        				return 1;
	        	}
        		else
        			BUG("checkForSpam() : Spam counter not found!");

		return 0;
       	}
	else
		delete rcs;

	return 0;
}

int RCQCache::handleCacheMissOnConnection(RCQPacket *p, RCParameters *rcpar,
					RCPattern *pat, int conID)
{
	RCEntry		*rce;
	RCRefList	*rl;
	RCQRPacket	*duplicate;
	int		eIndex;
	int		parIndex;

       	missedCacheCon++;

	/*
	 * Pass on, add parameters.
	 */
	rcpar->setSearchTTL(((RCQPacket *)p)->getTTL());
	rcpar->setInitTime(time(0) + startEntryInitTime);

	if ((parIndex = RCParametersDB->storeElement(rcpar)) == ENTRY_INVALID)
	{
		delete rcpar;
		cacheError++;
		BUG("handleCacheMissOnConnection() : Can't add parameters!");
		return 0;
	}

	rl	= new RCRefList(parIndex);
	if (!pat->addToRefList(RCPATTERN_REFLIST_PARAMETERS, rl))
	{
		cacheError++;
		BUG("handleCacheMissOnConnection() : Can't add "
				"reference to RCParameters!");
		delete rl;
		return 0;
	}

	/*
	 * Add entry
	 */
	rce	= new RCEntry(pat->getEntry(), conID);
	if ((eIndex = RCEntryDB->storeElement(rce)) == ENTRY_INVALID)
	{
		cacheError++;
		BUG("handleCacheMissOnConnection() : Adding Entry failed!");
		delete rce;
		RCParametersDB->deleteElement(parIndex);
		return 0;
	}

	pat->incRefCount();
	rl	= new RCRefList(eIndex);
	if (!pat->addToRefList(RCPATTERN_REFLIST_ENTRY, rl))
	{
		cacheError++;
		BUG("handleCacheMissOnConnection() : Can't add "
						"reference to Entry!");
		delete rl;
		return 0;
	}
	duplicate	= (RCQRPacket *)p->duplicate();
	duplicate->setTargetConnectionID(conID);
	this->addToOutList(duplicate);

	packetsSent	+= estimatePackets(((RCQPacket *)p)->getTTL());

	return 1;
}

int RCQCache::handleCacheHitOnConnection(RCQPacket *p, RCPattern *pat, int pIndex, int conID)
{
	RCEntry		*rce;
	RCQRPacket	*resp;
	RCQRPacket	*duplicate;
	RCQRPacket	*tmp;
	RCRefList	*rl;
	int		eIndex;
	int		hops;

	/*
	 * Update stats.
	 */
	packetsNotSent	+= estimatePackets(((RCQPacket *)p)->getTTL());
	hitOK++;

	/*
	 * Find correct entry
	 */
 	rce	= new RCEntry(pIndex, conID);
 	if ((eIndex = RCEntryDB->containsElement(rce)) == ENTRY_INVALID)
 	{
 		delete rce;
       		BUG("handleCacheHitOnConnection() : entry not found!");
		return 0;
 	}
 	delete rce;

 	rce	= (RCEntry *)RCEntryDB->findElement(eIndex);

 	/*
 	 * Determine whether there are packets to be returned
 	 */
	if (rce->getCounter() == COUNTER_ZERO)
 	{
 		hitZero++;
 		return 0;
 	}

 	if (rce->getCounter() == 0)
 	{
       		BUG("handleCacheHitOnConnection() : Zero packets!");
 		return 0;
 	}

//	if (RCStore->integrityCheck() < 0) *(int *)0=0;
 	resp	= (RCQRPacket *)RCStore->packetList(rce->getPacketID());
//	if (RCStore->integrityCheckByPacketID(rce->getPacketID()) < 0) *(int *)0=0;
	hops	= ((RCQPacket *)p)->getHops();
	for (rl = pat->getRefList(RCPATTERN_REFLIST_MESGID);
			rl; rl = rl->getNext())
	{
		RCMesgID	*target;

		if (!((target = (RCMesgID *)RCMesgIDDB->findElement(rl->getIndex()))))
		{
			BUG("handleCacheHitOnConnection() : MesgID not found!");
			return 0;
		}
		tmp	= resp;
//		if (RCStore->integrityCheckByPacketID(rce->getPacketID()) < 0) *(int *)0=0;
		while (tmp)
  		{
 			int	stl;

//			if (RCStore->integrityCheckByPacketID(rce->getPacketID()) < 0) *(int *)0=0;
 			if ((stl = resp->getFlushTime() - time(0)) > 0)
 			{
				duplicate	= (RCQRPacket *)tmp->duplicate();
				duplicate->setMesgID(target->getMesgIDp());
 				duplicate->setSTL(stl);
				duplicate->setTTL(hops);
				/*
				 * Doing this or not is a point of discussion
				 * should not be done when the result from a firewalled
				 * host, otherwise it is a good idea.
				 */
				// duplicate->setHops(7  - hops);
 				this->addToOutList(duplicate);
 			}
//			if (RCStore->integrityCheckByPacketID(rce->getPacketID()) < 0) *(int *)0=0;
 			tmp	= (RCQRPacket *)tmp->getNext();
			if (tmp == resp)
			{
				BUG("handleCacheHitOnConnection() : List has CIRCLE!");
//				if (RCStore->integrityCheck() < 0) *(int *)0=0;
//				*(int *)0=0;
				return 0;
			}
 		}
	}
//	if (RCStore->integrityCheck() < 0) *(int *)0=0;
	return 1;
}

int RCQCache::handleCacheHitInitTimeOnConnection(RCQPacket *p, RCMesgID *rcmid, RCPattern *pat)
{
	RCRefList	*rl;
	int 		mIndex;

	/*
	 * Update stats.
	 */
	hitInitTime++;
	packetsNotSent	+= estimatePackets(((RCQPacket *)p)->getTTL());

	if ((mIndex = RCMesgIDDB->storeElement(rcmid)) == ENTRY_INVALID)
	{
		BUG("handleCacheHitInitTimeOnConnection() : Adding mesgID failed!");

		mIndex = RCMesgIDDB->containsElement(rcmid);
		if (mIndex == ENTRY_INVALID)
		{
			delete rcmid;
			BUG("handleCacheHitInitTimeOnConnection() : present but not "
				"found!");
			return 0;
		}
		delete rcmid;
		if (!((rcmid = (RCMesgID *)RCMesgIDDB->findElement(mIndex))))
		{
			BUG("handleCacheHitInitTimeOnConnection() : can't fetch MegsID!");
			return 0;
		}
		if (rcmid->getPatternID() != pat->getEntry())
		{
			BUG("handleCacheHitInitTimeOnConnection() : indices don't match!");
			return 0;
		}
		else
			BUG("handleCacheHitInitTimeOnConnection() : double mesgID!");

		return 0;
	}
       	rl	= new RCRefList(mIndex);
       	if (!pat->addToRefList(RCPATTERN_REFLIST_MESGID, rl))
       	{
       		BUG("handleCacheHitInitTimeOnConnection() : add reference to mesgID "
			"failed!");
       		delete rl;
		return 0;
       	}
       	return 1;
}

int RCQCache::checkIntegrity()
{
	RCPattern	*p;
	int		i;

	for (i = 0; i < RCPatternDB->getSize(); i++)
	{
		RCRefList	*rl;
		RCMesgID	*rcmid;
		RCSpam		*rcs;
		RCEntry		*rce;
		RCParameters	*rcpar;
		int		refCount;
		int		realCount;
		int		j;

		if (!((p = (RCPattern *)RCPatternDB->findElement(i))))
			continue;

		/*
		 * MesgId's : check references
		 */
		refCount	= 0;
		realCount	= 0;
		for (rl = p->getRefList(RCPATTERN_REFLIST_MESGID); rl; rl = rl->getNext())
		{
			refCount++;
			if ((rcmid = (RCMesgID *)RCMesgIDDB->findElement(rl->getIndex())))
			{
				if (rcmid->getPatternID() != i)
				{
					BUG("RCQCache::checkIntegrity() : MesgID->getPatternID() wrong!");
					*(int *)0=0;
				}
			}
			else
			{
				BUG("RCQCache::checkIntegrity() : MesgID not found!");
				*(int *)0=0;
			}
		}

		/*
		 * count references
		 */
		for (j = 0; j < RCMesgIDDB->getSize(); j++)
			if ((rcmid = (RCMesgID *)RCMesgIDDB->findElement(j)))
				if (rcmid->getPatternID() == i)
					realCount++;
                              if (realCount != refCount)
		{
			BUG("RCQCache::checkIntegrity() : RCMesgID count wrong!");
			*(int *)0=0;
		}

		/*
		 * RCParameters
		 */
		refCount	= 0;
		realCount	= 0;
		for (rl = p->getRefList(RCPATTERN_REFLIST_PARAMETERS); rl; rl = rl->getNext())
   		{
			refCount++;
			if ((rcpar = (RCParameters *)RCParametersDB->findElement(rl->getIndex())))
			{
				if (rcpar->getPatternID() != i)
				{
					BUG("RCQCache::checkIntegrity() : RCParameters->getPatternID() wrong!");
					*(int *)0=0;
				}
			}
			else
			{
				BUG("RCQCache::checkIntegrity() : RCParameters not found!");
				*(int *)0=0;
			}
		}

		/*
		 * count references
		 */
		for (j = 0; j < RCParametersDB->getSize(); j++)
			if ((rcpar = (RCParameters *)RCParametersDB->findElement(j)))
				if (rcpar->getPatternID() == i)
					realCount++;
                              if (realCount != refCount)
		{
			BUG("RCQCache::checkIntegrity() : RCParameters count wrong!");
			*(int *)0=0;
		}

		/*
		 * RCSpam
		 */
		refCount	= 0;
		realCount	= 0;
		for (rl = p->getRefList(RCPATTERN_REFLIST_SPAM); rl; rl = rl->getNext())
 		{
			refCount++;
			if ((rcs = (RCSpam *)RCSpamDB->findElement(rl->getIndex())))
			{
				if (rcs->getPatternID() != i)
				{
					BUG("RCQCache::checkIntegrity() : RCSpam->getPatternID() wrong!");
					*(int *)0=0;
				}
			}
			else
			{
				BUG("RCQCache::checkIntegrity() : RCSpam not found!");
				*(int *)0=0;
			}
		}

		/*
		 * count references
		 */
		for (j = 0; j < RCSpamDB->getSize(); j++)
			if ((rcs = (RCSpam *)RCSpamDB->findElement(j)))
				if (rcs->getPatternID() == i)
					realCount++;
                              if (realCount != refCount)
		{
			BUG("RCQCache::checkIntegrity() : RCSpam count wrong!");
			*(int *)0=0;
		}

		/*
		 * RCSpam
		 */
		refCount	= 0;
		realCount	= 0;
		for (rl = p->getRefList(RCPATTERN_REFLIST_ENTRY); rl; rl = rl->getNext())
 		{
			refCount++;
			if ((rce = (RCEntry *)RCEntryDB->findElement(rl->getIndex())))
			{
				int	pktCounter;

				if (rce->getPatternID() != i)
				{
					BUG("RCQCache::checkIntegrity() : RCEntry->getPatternID() wrong!");
					*(int *)0=0;
				}
				if (rce->getCounter() == COUNTER_ZERO)
				{
					if (rce->getPacketID() != PKTID_INVALID)
					{
						BUG("RCQCache::checkIntegrity() : RCEntry->getPacketID() wrong!");
						*(int *)0=0;
					}
				}
				else
				{
					if (rce->getPacketID() == PKTID_INVALID)
					{
						BUG("RCQCache::checkIntegrity() : RCEntry->getPacketID() Invalid!");
						*(int *)0=0;
					}
					if ((pktCounter = RCStore->integrityCheckByPacketID(rce->getPacketID())) == INTEGRITY_ERROR)
					{
						BUG("RCQCache::checkIntegrity() : RCEntry : PacketStore Error!");
						*(int *)0=0;
					}
					else
					{
						if (pktCounter != rce->getCounter())
						{
							BUG("RCQCache::checkIntegrity() : RCEntry : Packet Count wrong!");
							*(int *)0=0;
						}
					}
				}
			}
			else
			{
				BUG("RCQCache::checkIntegrity() : RCEntry not found!");
				*(int *)0=0;
			}
		}

		/*
		 * count references
		 */
		for (j = 0; j < RCEntryDB->getSize(); j++)
			if ((rce = (RCEntry *)RCEntryDB->findElement(j)))
				if (rce->getPatternID() == i)
					realCount++;
                               if (realCount != refCount)
		{
			BUG("RCQCache::checkIntegrity() : RCEntries count wrong!");
			*(int *)0=0;
		}
	}

	return 1;
}
