/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "Config.h"

#if (OS == LINUX)
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#endif

#if (OS == FREEBSD)
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#endif

#include "Socket.h"

#if ((OS == LINUX) || (OS == FREEBSD) )
Socket::Socket()
{
	ip	= 0;
	port	= 0;
	sock	= -1;
}

Socket::Socket(Socket *s)
{
	ip	= s->getIp();
	port	= s->getPort();
	sock	= s->getSock();
}

Socket::Socket(int newSocket)
{
	ip	= 0;
	port	= 0;
	sock	= newSocket;
	if (fcntl(sock, F_SETFL, O_NONBLOCK) == -1)
		fprintf(stderr, "ERROR : Socket::Socket(int newSocket) : fcntl returned "
				"error!!!\r\n");
}

Socket::~Socket()
{
}

int Socket::callSocket()
{
         if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0)
		return 0;

	return (fcntl(sock, F_SETFL, O_NONBLOCK) != -1);
}

int Socket::callConnect(unsigned long hostIp, unsigned short hostPort)
{
	struct sockaddr_in	sa;

        ip			= hostIp;
        port			= hostPort;
        sa.sin_family   	= PF_INET;
        sa.sin_port     	= htons(port);
	sa.sin_addr.s_addr	= ip;
	
	if (connect(sock, (sockaddr *)&sa, sizeof(sa)) == -1)
	{
		if (errno == EINPROGRESS)
			return SOCKET_CONNECT_IN_PROGRESS;
		
		return SOCKET_CONNECT_FAILED;
	}
	
	return SOCKET_CONNECT_SUCCESS;
}

int Socket::testConnect()
{
	socklen_t	len;
	int		opt;
	
	if (sock == -1)
		return SOCKET_CONNECT_FAILED;
	
	if (!testSend(1))
		return SOCKET_CONNECT_IN_PROGRESS;
	
	len   = (socklen_t)sizeof(opt);
	if (getsockopt(sock, SOL_SOCKET, SO_ERROR, &opt, &len))
		return SOCKET_CONNECT_FAILED;
	
	if (opt)
		return SOCKET_CONNECT_FAILED;
	
	return SOCKET_CONNECT_SUCCESS;
}

int Socket::setSocket(int newSock)
{
	if (fcntl(sock, F_SETFL, O_NONBLOCK) == -1)
		return 0;

	if (sock != -1)
		this->callClose();

	sock	= newSock;

	return 1;
}

int Socket::callClose()
{
	int	r;

	r	= close(sock);
	sock	= -1;
	return r;
}

int Socket::testSend(int usec)
{
	fd_set  	out_set;
	struct timeval	tm;

	if (sock == -1)
	{
		fprintf(stderr, "ERROR : BUG : Socket::testSend() : testSend called with sock == -1!\r\n");
		return 0;
	}
	
	tm.tv_sec	= 0;
	tm.tv_usec	= usec;
	FD_ZERO(&out_set);
	FD_SET(sock, &out_set);

	return select(sock + 1, 0, &out_set, 0, &tm);
}

int Socket::testRecv(int usec)
{
	fd_set  		in_set;
	struct timeval	tm;

	if (sock == -1)
	{
		fprintf(stderr, "ERROR : BUG : Socket::testRecv() : testRecv called with sock == -1!\r\n");
		return 0;
	}
	
	tm.tv_sec	= 0;
	tm.tv_usec	= usec;
        FD_ZERO(&in_set);
        FD_SET(sock, &in_set);

	return select(sock + 1, &in_set, 0, 0, &tm);
}

int Socket::callRecv(char *buf, int count)
{
	return recv(sock, buf, count, 0);
}

int Socket::callSend(char *msg, int len)
{
	return send(sock, msg, len, 0);
}

int Socket::peek(char *buf, int count)
{
	return recv(sock, buf, count, MSG_PEEK);
}

int Socket::callBind(unsigned short port)
{
	const int		reuseaddr = 1;
	struct sockaddr_in	sa;

	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuseaddr,
			sizeof(reuseaddr));

	sa.sin_family		= PF_INET;
	sa.sin_addr.s_addr	= htonl(INADDR_ANY);
	sa.sin_port     	= htons(port);

	return bind(sock, (struct sockaddr *)&sa, sizeof(sa));
}

int Socket::callListen(int queue)
{
	return listen(sock, queue);
}

Socket *Socket::callAccept()
{
	Socket			*ns;
	struct sockaddr_in	sa;
	int			s;
	unsigned int		len;

	len	= sizeof(sa);
	if ((s = accept(sock, (struct sockaddr *)&sa, &len)) < 0)
		return 0;

	ns	= new Socket(s);
	ns->setIp(sa.sin_addr.s_addr);
	ns->setPort(sa.sin_port);
	
	return ns;
}

void Socket::getDottedIp(char *buf)
{
	struct in_addr	ia;
	
	ia.s_addr	= ip;
	sprintf(buf, "%s", inet_ntoa(ia));
}

#endif

int Socket::getSock()
{
	return sock;
}

unsigned long int Socket::getIp()
{
	return ip;
}

unsigned short Socket::getPort()
{
	return port;
}

void Socket::setIp(unsigned long int newIp)
{
	ip	= newIp;
}

void Socket::setPort(unsigned short newPort)
{
	port	= newPort;
}

